"""Tests for dials.merge command line program."""


from __future__ import annotations

import json

import procrunner
import pytest

from cctbx import uctbx
from dxtbx.serialize import load
from iotbx import mtz

from dials.array_family import flex


def validate_mtz(mtz_file, expected_labels, unexpected_labels=None):
    assert mtz_file.is_file()
    m = mtz.object(str(mtz_file))

    assert m.as_miller_arrays()[0].info().wavelength == pytest.approx(0.6889)
    labels = set()
    for ma in m.as_miller_arrays(merge_equivalents=False):
        labels.update(ma.info().labels)
    for l in expected_labels:
        assert l in labels
    if unexpected_labels:
        for l in unexpected_labels:
            assert l not in labels


@pytest.mark.parametrize("anomalous", [True, False])
@pytest.mark.parametrize(
    "truncate,french_wilson_impl", [(True, "dials"), (True, "cctbx"), (False, None)]
)
def test_merge(dials_data, tmp_path, anomalous, truncate, french_wilson_impl):
    """Test the command line script with LCY data"""
    # Main options: truncate on/off, anomalous on/off
    # french_wilson.implementation dials/cctbx

    mean_labels = ["IMEAN", "SIGIMEAN"]
    anom_labels = ["I(+)", "I(-)", "SIGI(+)", "SIGI(-)"]
    amp_labels = ["F", "SIGF"]
    anom_amp_labels = ["F(+)", "SIGF(+)", "F(-)", "SIGF(-)", "DANO", "SIGDANO"]

    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refls = location / "scaled_20_25.refl"
    expts = location / "scaled_20_25.expt"

    mtz_file = tmp_path / f"merge-{anomalous}-{truncate}.mtz"

    command = [
        "dials.merge",
        refls,
        expts,
        f"truncate={truncate}",
        f"french_wilson.implementation={french_wilson_impl}",
        f"anomalous={anomalous}",
        f"output.mtz={str(mtz_file)}",
        "project_name=ham",
        "crystal_name=jam",
        "dataset_name=spam",
        "json=dials.merge.json",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "dials.merge.html").is_file()
    merge_json = tmp_path / "dials.merge.json"
    assert merge_json.is_file()
    expected_labels = mean_labels
    unexpected_labels = []

    with merge_json.open() as fh:
        json_d = json.load(fh)
        wl = list(json_d.keys())[0]
        for k in {"merging_stats", "merging_stats_anom"}:
            assert k in json_d[wl]
            assert {"d_star_sq_min", "n_obs", "cc_anom"} <= json_d[wl][k].keys()

    if truncate:
        expected_labels += amp_labels
    else:
        unexpected_labels += amp_labels

    if anomalous:
        expected_labels += anom_labels
    else:
        unexpected_labels += anom_labels

    if anomalous and truncate:
        expected_labels += anom_amp_labels
    else:
        unexpected_labels += anom_amp_labels

    validate_mtz(mtz_file, expected_labels, unexpected_labels)


@pytest.mark.parametrize("best_unit_cell", [None, "5.5,8.1,12.0,90,90,90"])
def test_merge_dmin_dmax(dials_data, tmp_path, best_unit_cell):
    """Test the d_min, d_max"""

    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refls = location / "scaled_20_25.refl"
    expts = location / "scaled_20_25.expt"

    mtz_file = tmp_path / "merge.mtz"

    command = [
        "dials.merge",
        refls,
        expts,
        "truncate=False",
        "anomalous=False",
        "d_min=1.0",
        "d_max=8.0",
        f"output.mtz={str(mtz_file)}",
        "project_name=ham",
        "crystal_name=jam",
        "dataset_name=spam",
        f"best_unit_cell={best_unit_cell}",
        "output.html=None",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    # check the unit cell was correctly set if using best_unit_cell
    m = mtz.object(str(mtz_file))
    if best_unit_cell:
        for ma in m.as_miller_arrays():
            assert uctbx.unit_cell(best_unit_cell).parameters() == pytest.approx(
                ma.unit_cell().parameters()
            )

    # check we only have reflections in range 8 - 1A
    max_min_resolution = m.max_min_resolution()

    assert max_min_resolution[0] <= 8
    assert max_min_resolution[1] >= 1


def test_merge_multi_wavelength(dials_data, tmp_path):
    """Test that merge handles multi-wavelength data suitably - should be
    exported into an mtz with separate columns for each wavelength."""

    mean_labels = [f"{pre}IMEAN_WAVE{i}" for i in [1, 2] for pre in ["", "SIG"]]
    anom_labels = [
        f"{pre}I_WAVE{i}({sgn})"
        for i in [1, 2]
        for pre in ["", "SIG"]
        for sgn in ["+", "-"]
    ]
    amp_labels = [f"{pre}F_WAVE{i}" for i in [1, 2] for pre in ["", "SIG"]]
    anom_amp_labels = [
        f"{pre}F_WAVE{i}({sgn})"
        for i in [1, 2]
        for pre in ["", "SIG"]
        for sgn in ["+", "-"]
    ]

    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refl1 = location / "scaled_30.refl"
    expt1 = location / "scaled_30.expt"
    refl2 = location / "scaled_35.refl"
    expt2 = location / "scaled_35.expt"
    expts1 = load.experiment_list(expt1, check_format=False)
    expts1[0].beam.set_wavelength(0.7)
    expts2 = load.experiment_list(expt2, check_format=False)
    expts1.extend(expts2)

    tmp_expt = tmp_path / "tmp.expt"
    expts1.as_json(tmp_expt)

    reflections1 = flex.reflection_table.from_file(refl1)
    reflections2 = flex.reflection_table.from_file(refl2)
    # first need to resolve identifiers - usually done on loading
    reflections2["id"] = flex.int(reflections2.size(), 1)
    del reflections2.experiment_identifiers()[0]
    reflections2.experiment_identifiers()[1] = "3"
    reflections1.extend(reflections2)

    tmp_refl = tmp_path / "tmp.refl"
    reflections1.as_file(tmp_refl)

    # Can now run after creating our 'fake' multiwavelength dataset
    command = ["dials.merge", tmp_refl, tmp_expt, "truncate=True", "anomalous=True"]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "merged.mtz").is_file()
    assert (tmp_path / "dials.merge.html").is_file()
    m = mtz.object(str(tmp_path / "merged.mtz"))
    labels = []
    for ma in m.as_miller_arrays(merge_equivalents=False):
        labels.extend(ma.info().labels)
    assert all(x in labels for x in mean_labels)
    assert all(x in labels for x in anom_labels)
    assert all(x in labels for x in amp_labels)
    assert all(x in labels for x in anom_amp_labels)

    # 6 miller arrays for each dataset, check the expected number of reflections.
    arrays = m.as_miller_arrays()
    assert len(arrays) == 12
    assert arrays[0].info().wavelength == pytest.approx(0.7)
    assert arrays[6].info().wavelength == pytest.approx(0.6889)
    assert abs(arrays[0].size() - 1223) < 10  # check number of miller indices
    assert abs(arrays[6].size() - 1453) < 10  # check number of miller indices

    # test changing the wavelength tolerance such that data is combined under
    # one wavelength. Check the number of reflections to confirm this.
    command = [
        "dials.merge",
        tmp_refl,
        tmp_expt,
        "truncate=True",
        "anomalous=True",
        "wavelength_tolerance=0.02",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    m = mtz.object(str(tmp_path / "merged.mtz"))
    arrays = m.as_miller_arrays()
    assert arrays[0].info().wavelength == pytest.approx(0.7)
    assert len(arrays) == 6
    assert abs(arrays[0].size() - 1538) < 10


def test_suitable_exit_for_bad_input_from_single_dataset(dials_data, tmp_path):
    location = dials_data("vmxi_proteinase_k_sweeps", pathlib=True)

    command = [
        "dials.merge",
        location / "experiments_0.json",
        location / "reflections_0.pickle",
    ]

    # unscaled data
    result = procrunner.run(command, working_directory=tmp_path)
    assert result.returncode
    assert (
        result.stderr.replace(b"\r", b"")
        == b"""Sorry: intensity.scale.value not found in the reflection table.
Only scaled data can be processed with dials.merge
"""
    )


def test_suitable_exit_for_bad_input_with_more_than_one_reflection_table(
    dials_data, tmp_path
):
    location = dials_data("vmxi_proteinase_k_sweeps", pathlib=True)

    command = [
        "dials.merge",
        location / "experiments_0.json",
        location / "reflections_0.pickle",
        location / "experiments_1.json",
        location / "reflections_1.pickle",
    ]

    # more than one reflection table.
    result = procrunner.run(command, working_directory=tmp_path)
    assert result.returncode
    assert (
        result.stderr.replace(b"\r", b"")
        == b"""Sorry: Only data scaled together as a single reflection dataset
can be processed with dials.merge
"""
    )



from __future__ import annotations

geometry_phil = """geometry.parameters
    .help = "Set up an experimental model for refinement test case"
{
    random_seed = 2
        .type = int
    beam
    {
        wavelength
        {
            random = True
                .type = bool
            range = 0.8 1.2
                .type = floats(size=2)
            value = None
                .type = float
        }
        direction
        {
            method = *inclination close_to exactly
                .type = choice
            inclination
                .help = "inclination in the X-Z plane"
            {
                random = True
                    .type = bool
                angle = 0.5
                    .type = float
            }
            close_to
            {
                direction = None
                    .type = floats(size=3)
                sd = 0.5
                    .type = float
            }
            exactly = None
                .type = floats(size=3)
        }
    }
    crystal
    {
        space_group_symbol = "P 1"
            .type = str
        a
        {
            length
            {
                random = True
                    .type = bool
                range = 10 20
                    .type = floats(size=2)
                value = None
                    .type = float
            }
            direction
            {
                method = *close_to exactly
                    .type = choice
                close_to
                {
                        direction = 1., 0., 0.
                            .type = floats(size=3)
                        sd = 0.5
                            .type = float
                }
                exactly
                {
                        direction = 1., 0., 0.
                            .type = floats(size=3)
                }
            }


        }
        b
        {
            length
            {
                random = True
                    .type = bool
                range = 10 20
                    .type = floats(size=2)
                value = None
                    .type = float
            }
            direction
            {
                method = *close_to exactly
                    .type = choice
                close_to
                {
                        direction = 0., 1., 0.
                            .type = floats(size=3)
                        sd = 0.5
                            .type = float
                }
                exactly
                {
                        direction = 0., 1., 0.
                            .type = floats(size=3)
                }
            }


        }
        c
        {
            length
            {
                random = True
                    .type = bool
                range = 10 20
                    .type = floats(size=2)
                value = None
                    .type = float
            }
            direction
            {
                method = *close_to exactly
                    .type = choice
                close_to
                {
                        direction = 0., 0., 1.
                            .type = floats(size=3)
                        sd = 0.5
                            .type = float
                }
                exactly
                {
                        direction = 0., 0., 1.
                            .type = floats(size=3)
                }
            }


        }
    }
    goniometer
    {
        axis = 1. 0. 0.
            .type = floats(size=3)
    }
    detector
    {
        directions
        {
            method = *close_to exactly
                .type=choice
            close_to
            {
                dir1 = 1., 0., 0.
                    .type = floats(size=3)
                norm = 0., 0., 1.
                    .type = floats(size=3)
                sd = 0.5
                    .type = float
            }
            exactly
            {
                dir1 = 1., 0., 0.
                    .type = floats(size=3)
                norm = 0., 0., 1.
                    .type = floats(size=3)
            }
        }
        centre
        {
            method = *close_to exactly
                .type=choice
            close_to
            {
                value = 0., 0., 200.
                    .type = floats(size=3)
                sd = 0.5
                    .type = float
            }
            exactly
            {
                value = 0., 0., 200.
                    .type = floats(size=3)
            }
        }
        npx_fast = 1475
            .type = int
        npx_slow = 1679
            .type = int
        pix_size = 0.172
            .type = float
    }
}"""


minimiser_phil = """minimiser.parameters
    .help = "Set up an minimiser for refinement test case"
{
    engine = SimpleLBFGS LBFGScurvs *GaussNewton
        .type = choice
    logfile = tst_orientation_refinement.log
        .type = path
}"""



from __future__ import annotations

from unittest import mock

import dials.util.phil
from dials.array_family import flex


@mock.patch("dials.util.phil.ExperimentListFactory")
def test(ExperimentListFactory, dials_data):
    # Only use these filenames for verification
    path = dials_data("centroid_test_data", pathlib=True)
    experiments_path = path / "experiments.json"
    reflections_path1 = path / "integrated.pickle"
    reflections_path2 = path / "integrated.refl"

    phil_scope = dials.util.phil.parse(
        f"""
    input {{
      reflections1 = {reflections_path1}
        .type = reflection_table
      reflections2 = {reflections_path2}
        .type = reflection_table
      experiments = {experiments_path}
        .type = experiment_list
    }}
  """
    )

    params = phil_scope.extract()

    # Check the right filenames were parsed
    assert params.input.reflections1.filename == str(reflections_path1)
    assert params.input.reflections2.filename == str(reflections_path2)
    assert params.input.experiments.filename == str(experiments_path)
    # Check that we got the expected objects back
    assert isinstance(params.input.experiments.data, mock.Mock)
    assert isinstance(params.input.reflections1.data, flex.reflection_table)
    assert isinstance(params.input.reflections2.data, flex.reflection_table)
    # Check we had the correct calls made
    assert ExperimentListFactory.from_json_file.call_args[0] == (str(experiments_path),)



from __future__ import annotations

import dials.algorithms.profile_model.modeller  # noqa: F401; lgtm; true import dependency
from dials.algorithms.profile_model.gaussian_rs.model import Model, phil_scope
from dials_algorithms_profile_model_gaussian_rs_ext import (
    BBoxCalculator2D,
    BBoxCalculator3D,
    BBoxCalculatorIface,
    BBoxMultiCalculator,
    CoordinateSystem,
    CoordinateSystem2d,
    GaussianRSProfileModeller,
    MaskCalculator2D,
    MaskCalculator3D,
    MaskCalculatorIface,
    MaskMultiCalculator,
    PartialityCalculator2D,
    PartialityCalculator3D,
    PartialityCalculatorIface,
    PartialityMultiCalculator,
    ideal_profile_double,
    ideal_profile_float,
    zeta_factor,
)

__all__ = [
    "BBoxCalculator",
    "BBoxCalculator2D",
    "BBoxCalculator3D",
    "BBoxCalculatorIface",
    "BBoxMultiCalculator",
    "CoordinateSystem",
    "CoordinateSystem2d",
    "GaussianRSProfileModeller",
    "MaskCalculator",
    "MaskCalculator2D",
    "MaskCalculator3D",
    "MaskCalculatorIface",
    "MaskMultiCalculator",
    "Model",
    "PartialityCalculator",
    "PartialityCalculator2D",
    "PartialityCalculator3D",
    "PartialityCalculatorIface",
    "PartialityMultiCalculator",
    "ideal_profile_double",
    "ideal_profile_float",
    "phil_scope",
    "zeta_factor",
]


def BBoxCalculator(crystal, beam, detector, goniometer, scan, delta_b, delta_m):
    """Return the relevant bbox calculator."""
    if goniometer is None or scan is None or scan.is_still():
        algorithm = BBoxCalculator2D(beam, detector, delta_b, delta_m)
    else:
        algorithm = BBoxCalculator3D(beam, detector, goniometer, scan, delta_b, delta_m)
    return algorithm


def PartialityCalculator(crystal, beam, detector, goniometer, scan, sigma_m):
    """Return the relevant partiality calculator."""
    if goniometer is None or scan is None or scan.is_still():
        algorithm = PartialityCalculator2D(beam, sigma_m)
    else:
        algorithm = PartialityCalculator3D(beam, goniometer, scan, sigma_m)
    return algorithm


def MaskCalculator(crystal, beam, detector, goniometer, scan, delta_b, delta_m):
    """Return the relevant partiality calculator."""
    if goniometer is None or scan is None or scan.is_still():
        algorithm = MaskCalculator2D(beam, detector, delta_b, delta_m)
    else:
        algorithm = MaskCalculator3D(beam, detector, goniometer, scan, delta_b, delta_m)
    return algorithm



from __future__ import annotations

import os

import procrunner
import pytest


def test_basic(dials_regression, run_in_tmp_path):
    # Call dials.create_profile_model
    result = procrunner.run(
        [
            "dials.create_profile_model",
            os.path.join(
                dials_regression,
                "integration_test_data",
                "i04-weak-data2",
                "experiments.json",
            ),
            os.path.join(
                dials_regression,
                "integration_test_data",
                "i04-weak-data2",
                "indexed.pickle",
            ),
            "sigma_m_algorithm=basic",
        ]
    )
    assert not result.returncode and not result.stderr
    assert os.path.exists("models_with_profiles.expt")

    from dxtbx.model.experiment_list import ExperimentListFactory

    experiments = ExperimentListFactory.from_json_file(
        "models_with_profiles.expt", check_format=False
    )
    sigma_b = experiments[0].profile.sigma_b(deg=True)
    sigma_m = experiments[0].profile.sigma_m(deg=True)
    assert sigma_b == pytest.approx(0.02446, abs=1e-3)
    assert sigma_m == pytest.approx(0.06833, abs=1e-3)


def test_extended(dials_regression, run_in_tmp_path):
    # Call dials.create_profile_model
    result = procrunner.run(
        [
            "dials.create_profile_model",
            os.path.join(
                dials_regression,
                "integration_test_data",
                "i04-weak-data2",
                "experiments.json",
            ),
            os.path.join(
                dials_regression,
                "integration_test_data",
                "i04-weak-data2",
                "indexed.pickle",
            ),
        ]
    )
    assert not result.returncode and not result.stderr
    assert os.path.exists("models_with_profiles.expt")

    from dxtbx.model.experiment_list import ExperimentListFactory

    experiments = ExperimentListFactory.from_json_file(
        "models_with_profiles.expt", check_format=False
    )
    sigma_b = experiments[0].profile.sigma_b(deg=True)
    sigma_m = experiments[0].profile.sigma_m(deg=True)
    assert sigma_b == pytest.approx(0.02446, abs=1e-3)
    assert sigma_m == pytest.approx(0.04187, abs=1e-3)



"""
Command line script to allow merging and truncating of a dials dataset.
"""

from __future__ import annotations

import json
import logging
import sys
from contextlib import contextmanager
from io import StringIO

from dxtbx.model import ExperimentList
from iotbx import phil

from dials.algorithms.merging.merge import (
    MTZDataClass,
    make_merged_mtz_file,
    merge,
    show_wilson_scaling_analysis,
    truncate,
)
from dials.algorithms.merging.reporting import (
    MergeJSONCollector,
    generate_html_report,
    make_dano_table,
)
from dials.algorithms.scaling.scaling_library import determine_best_unit_cell
from dials.util import Sorry, log, show_mail_handle_errors
from dials.util.export_mtz import match_wavelengths
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

help_message = """
Merge scaled dials data.

Examples::

  dials.merge scaled.expt scaled.refl

  dials.merge scaled.expt scaled.refl truncate=False
"""

logger = logging.getLogger("dials")
phil_scope = phil.parse(
    """
assess_space_group = True
    .type = bool
    .help = "Option to assess space group by testing presence of axial reflections"
anomalous = True
    .type = bool
    .help = "Output anomalous as well as mean intensities."
truncate = True
    .type = bool
    .help = "Use the French & Wilson (1978) algorithm to correct for negative "
            "intensities when estimating amplitudes."
french_wilson {
    implementation = *dials cctbx
        .type = choice
        .help = "Choice of implementation of the French & Wilson algorithm"
    min_reflections = 200
        .type = int(value_min=1)
        .help = "Only perform French & Wilson procedure if at least this "
                "number of reflections."
    fallback_to_flat_prior = True
        .type = bool
        .help = "If insufficient number of reflections to perform the "
                "French & Wilson procedure, fallback to assumption of a "
                "flat prior, i.e.: "
                "  |F| = sqrt((Io+sqrt(Io**2 +2sigma**2))/2.0)"
}
d_min = None
    .type = float
    .help = "High resolution limit to apply to the data."
d_max = None
    .type = float
    .help = "Low resolution limit to apply to the data."
wavelength_tolerance = 1e-4
    .type = float(value_min=0.0)
    .help = "Absolute tolerance for determining wavelength grouping for merging."
combine_partials = True
    .type = bool
    .help = "Combine partials that have the same partial id into one
        reflection, with an updated partiality given by the sum of the
        individual partialities."
partiality_threshold=0.4
    .type = float
    .help = "All reflections with partiality values above the partiality
        threshold will be retained. This is done after any combination of
        partials if applicable."
best_unit_cell = None
    .type = unit_cell
    .help = "Best unit cell value, to use when performing resolution cutting,"
            "and as the overall unit cell in the merged mtz. If undefined, the median"
            "cell will be used."
n_residues = 200
    .type = int
    .help = "Number of residues to use in Wilson scaling"
merging {
    use_internal_variance = False
        .type = bool
    n_bins = 20
        .type = int(value_min=5)
    anomalous = False
        .type = bool
        .help = "Option to control whether reported merging stats are anomalous."
}
output {
    log = dials.merge.log
        .type = str
    mtz = merged.mtz
        .type = str
        .help = "Filename to use for mtz output."
    html = dials.merge.html
        .type = str
        .help = "Filename for html output report."
    json = None
        .type = str
        .help = "Filename to output data from html report in json format."
    crystal_names = XTAL
        .type = strings
        .help = "Crystal name to be used in MTZ file output (multiple names
            allowed for MAD datasets)"
    project_name = AUTOMATIC
        .type = str
        .help = "Project name to be used in MTZ file output"
    dataset_names = NATIVE
        .type = strings
        .help = "Dataset name to be used in MTZ file output (multiple names
            allowed for MAD datasets)"
}
""",
    process_includes=True,
)


@contextmanager
def collect_html_data_from_merge():
    try:
        html_collector = MergeJSONCollector()
        yield html_collector
    finally:
        html_collector.reset()


def merge_data_to_mtz_with_report_collection(params, experiments, reflections):
    """Run the merge_data_to_mtz function, also collecting data for json/html output"""
    with collect_html_data_from_merge() as collector:
        mtz = merge_data_to_mtz(params, experiments, reflections)
        json_data = collector.create_json()
    return mtz, json_data


def merge_data_to_mtz(params, experiments, reflections):
    """Merge data (at each wavelength) and write to an mtz file object."""
    wavelengths = match_wavelengths(
        experiments,
        absolute_tolerance=params.wavelength_tolerance,
    )  # wavelengths is an ordered dict
    mtz_datasets = [
        MTZDataClass(wavelength=w, project_name=params.output.project_name)
        for w in wavelengths.keys()
    ]
    dataset_names = params.output.dataset_names
    crystal_names = params.output.crystal_names

    # check if best_unit_cell is set.
    best_unit_cell = params.best_unit_cell
    if not best_unit_cell:
        best_unit_cell = determine_best_unit_cell(experiments)
    reflections[0]["d"] = best_unit_cell.d(reflections[0]["miller_index"])
    for expt in experiments:
        expt.crystal.unit_cell = best_unit_cell

    if len(wavelengths) > 1:
        logger.info(
            "Multiple wavelengths found: \n%s",
            "\n".join(
                "  Wavlength: %.5f, experiment numbers: %s "
                % (k, ",".join(map(str, v)))
                for k, v in wavelengths.items()
            ),
        )
        if not dataset_names or len(dataset_names) != len(wavelengths):
            logger.info(
                "Unequal number of dataset names and wavelengths, using default naming."
            )
            dataset_names = [None] * len(wavelengths)
        if not crystal_names or len(crystal_names) != len(wavelengths):
            logger.info(
                "Unequal number of crystal names and wavelengths, using default naming."
            )
            crystal_names = [None] * len(wavelengths)
        experiments_subsets = []
        reflections_subsets = []
        for dataset, dname, cname in zip(mtz_datasets, dataset_names, crystal_names):
            dataset.dataset_name = dname
            dataset.crystal_name = cname
        for exp_nos in wavelengths.values():
            expids = [experiments[i].identifier for i in exp_nos]
            experiments_subsets.append(
                ExperimentList([experiments[i] for i in exp_nos])
            )
            reflections_subsets.append(
                reflections[0].select_on_experiment_identifiers(expids)
            )
    else:
        mtz_datasets[0].dataset_name = dataset_names[0]
        mtz_datasets[0].crystal_name = crystal_names[0]
        experiments_subsets = [experiments]
        reflections_subsets = reflections

    # merge and truncate the data for each wavelength group
    for experimentlist, reflection_table, mtz_dataset in zip(
        experiments_subsets, reflections_subsets, mtz_datasets
    ):
        # First generate two merge_equivalents objects, collect merging stats
        merged, merged_anomalous, stats_summary = merge(
            experimentlist,
            reflection_table,
            d_min=params.d_min,
            d_max=params.d_max,
            combine_partials=params.combine_partials,
            partiality_threshold=params.partiality_threshold,
            best_unit_cell=best_unit_cell,
            anomalous=params.anomalous,
            assess_space_group=params.assess_space_group,
            n_bins=params.merging.n_bins,
            use_internal_variance=params.merging.use_internal_variance,
        )

        merged_array = merged.array()
        # Save the relevant data in the mtz_dataset dataclass
        # This will add the data for IMEAN/SIGIMEAN
        mtz_dataset.merged_array = merged_array
        if merged_anomalous:
            merged_anomalous_array = merged_anomalous.array()
            # This will add the data for I(+), I(-), SIGI(+), SIGI(-), N(+), N(-)
            mtz_dataset.merged_anomalous_array = merged_anomalous_array
            mtz_dataset.multiplicities = merged_anomalous.redundancies()
        else:
            merged_anomalous_array = None
            # This will add the data for N
            mtz_dataset.multiplicities = merged.redundancies()

        if params.anomalous:
            merged_intensities = merged_anomalous_array
        else:
            merged_intensities = merged_array

        anom_amplitudes = None
        if params.truncate:
            amplitudes, anom_amplitudes, dano = truncate(
                merged_intensities,
                implementation=params.french_wilson.implementation,
                min_reflections=params.french_wilson.min_reflections,
                fallback_to_flat_prior=params.french_wilson.fallback_to_flat_prior,
            )
            # This will add the data for F, SIGF
            mtz_dataset.amplitudes = amplitudes
            # This will add the data for F(+), F(-), SIGF(+), SIGF(-)
            mtz_dataset.anomalous_amplitudes = anom_amplitudes
            # This will add the data for DANO, SIGDANO
            mtz_dataset.dano = dano

        # print out analysis statistics
        B_iso = show_wilson_scaling_analysis(merged_intensities)
        stats_summary.Wilson_B_iso = B_iso

        if anom_amplitudes:
            logger.info(make_dano_table(anom_amplitudes))

        if stats_summary.merging_statistics_result:
            logger.info(stats_summary)

        if MergeJSONCollector.initiated:
            stats_summary.anomalous_amplitudes = anom_amplitudes
            MergeJSONCollector.data[mtz_dataset.wavelength] = stats_summary

    # pass the dataclasses to an MTZ writer to generate the mtz file and return.
    return make_merged_mtz_file(mtz_datasets)


@show_mail_handle_errors()
def run(args=None):
    """Run the merging from the command-line."""
    usage = """Usage: dials.merge scaled.refl scaled.expt [options]"""

    parser = ArgumentParser(
        usage=usage,
        read_experiments=True,
        read_reflections=True,
        phil=phil_scope,
        check_format=False,
        epilog=help_message,
    )
    params, options = parser.parse_args(args=args, show_diff_phil=False)

    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        sys.exit()

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    log.config(verbosity=options.verbose, logfile=params.output.log)
    logger.info(dials_version())

    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    ### Assert that all data have been scaled with dials - should only be
    # able to input one reflection table and experimentlist that are
    # matching and scaled together.

    if len(reflections) != 1:
        raise Sorry(
            """Only data scaled together as a single reflection dataset
can be processed with dials.merge"""
        )

    for k in [
        "intensity.scale.value",
        "intensity.scale.variance",
        "inverse_scale_factor",
    ]:
        if k not in reflections[0]:
            raise Sorry(
                f"""{k} not found in the reflection table.
Only scaled data can be processed with dials.merge"""
            )

    try:
        if params.output.json or params.output.html:
            mtz_file, json_data = merge_data_to_mtz_with_report_collection(
                params, experiments, reflections
            )
        else:
            mtz_file = merge_data_to_mtz(params, experiments, reflections)
            json_data = {}
    except ValueError as e:
        raise Sorry(e)

    logger.info("\nWriting reflections to %s", (params.output.mtz))
    out = StringIO()
    mtz_file.show_summary(out=out)
    logger.info(out.getvalue())
    mtz_file.write(params.output.mtz)

    if params.output.json:
        with open(params.output.json, "w", encoding="utf-8") as f:
            json.dump(json_data, f, indent=2)
    if params.output.html:
        generate_html_report(json_data, params.output.html)


if __name__ == "__main__":
    run()



from __future__ import annotations

import pytest


@pytest.fixture
def prediction():
    from dials.model.data import Prediction

    pred = Prediction()
    pred.miller_index = (1, 2, 3)
    pred.beam_vector = (4, 5, 6)
    pred.panel = 7
    pred.entering = True
    pred.position.px = (8, 9, 10)
    pred.position.mm = (11, 12, 13)
    return pred


def test_data(prediction):
    eps = 1e-7

    assert prediction.miller_index == (1, 2, 3)
    assert prediction.panel == 7
    assert prediction.entering
    assert abs(prediction.beam_vector[0] - 4) <= eps
    assert abs(prediction.beam_vector[1] - 5) <= eps
    assert abs(prediction.beam_vector[2] - 6) <= eps
    assert abs(prediction.position.px[0] - 8) <= eps
    assert abs(prediction.position.px[1] - 9) <= eps
    assert abs(prediction.position.px[2] - 10) <= eps
    assert abs(prediction.position.mm[0] - 11) <= eps
    assert abs(prediction.position.mm[1] - 12) <= eps
    assert abs(prediction.position.mm[2] - 13) <= eps


def test_equality(prediction):
    from dials.model.data import Prediction

    pred2 = Prediction(prediction)
    assert pred2 == prediction
    pred2 = Prediction(prediction)
    pred2.miller_index = (0, 0, 0)
    assert pred2 != prediction
    pred2 = Prediction(prediction)
    pred2.beam_vector = (0, 0, 0)
    assert pred2 != prediction
    pred2 = Prediction(prediction)
    pred2.panel = 0
    assert pred2 != prediction
    pred2 = Prediction(prediction)
    pred2.entering = False
    assert pred2 != prediction
    pred2 = Prediction(prediction)
    pred2.position.px = (0, 0, 0)
    assert pred2 != prediction
    pred2 = Prediction(prediction)
    pred2.position.mm = (0, 0, 0)
    assert pred2 != prediction



from __future__ import annotations

from libtbx.phil import command_line

from dials.algorithms.simulation.generate_test_reflections import main, master_phil


def test_generate_test_reflections(run_in_tmp_path):
    cmd = command_line.argument_interpreter(master_phil=master_phil)
    working_phil = cmd.process_and_fetch(
        args=[
            """nrefl = 10
shoebox_size {
  x = 20
  y = 20
  z = 20
}
spot_size {
  x = 1
  y = 3
  z = 1
}
spot_offset {
  x = -0.5
  y = -0.5
  z = -0.5
}
mask_nsigma = 3.0
counts = 10000
background = 0
background_a = 10
background_b = 0.1
background_c = -0.1
background_d = 0
pixel_mask = *all static precise
background_method = *xds mosflm
integration_methpd = *xds mosflm
output {
  over = None
  under = None
  all = all_refl.refl
}
rotation {
  axis {
    x = 0
    y = 0
    z = 1
  }
  angle = 45
}
"""
        ]
    )
    main(working_phil.extract())
    assert (run_in_tmp_path / "all_refl.refl").exists()



"""
Use a powder diffraction pattern of a known standard to calibrate goniometer geometry.
This can be either an electron or X-ray diffraction pattern. The parameters
calibrated are beam position, beam distance and beam wavelength. Detector tilt
calibration can be implemented as need arises.

This code wraps [pyFAI](https://doi.org/10.1107/S1600576715004306),
a well established X-ray powder diffraction tool.

The calibration is done in two steps:
    Step 1) A coarse beam position calibration done by the user using a matplolib widget.
    Step 2) Starting from the coarse calibration, pyFAI geometry calibration provides
a fine full geometry calibration.


Usage examples:
1. Use the widget to generate a starting, coarse geometry after which the fine
geometry calibration is applied. Overwrite the default name of the coarse geometry output file
    $ dials.powder_calibrate poor_geom.expt standard="Al" coarse_geom="a_starting_geom.expt"

2. Starting from a coarse geometry file, perform fine calibration with pyFAI bypassing the
widget tool.
    $ dials.powder_calibrate coarse_geom.expt standard="Al" eyeball=False
"""

from __future__ import annotations

import logging
import os
from sys import exit
from typing import List, NamedTuple, Optional, Tuple, Union

import numpy as np
import numpy.ma as ma
from matplotlib import pyplot as plt
from matplotlib.widgets import Button, Slider

try:
    import pyFAI  # noqa: F401
except ModuleNotFoundError:
    exit(
        """
            This script requires the pyFAI library. Try:
                $ conda install -c conda-forge pyFAI-base
                   OR if you prefer mamba
                $ mamba install pyFAI-base
            """
    )
from pyFAI.calibrant import get_calibrant as pfCalibrant
from pyFAI.detectors import Detector as pfDetector
from pyFAI.geometry import Geometry as pfGeometry
from pyFAI.goniometer import SingleGeometry as pfSingleGeometry
from pyFAI.gui import jupyter as pfjupyter

from dxtbx.model import ExperimentList
from libtbx.phil import parse, scope, scope_extract

# dxtbx and dials must be imported after pyfai,
# the alternative causes pyfai to segv
import dials.util
import dials.util.log
from dials.command_line.modify_geometry import update as geometry_update
from dials.util.options import ArgumentParser, flatten_experiments
from dials.util.version import dials_version

# Define a logger.
logger = logging.getLogger("dials.command_line.powder_calibrate")

phil_scope = parse(
    """
    standard = None
        .type = str
        .help = calibrant name
    eyeball = True
        .type = bool
        .help = use widget to eyeball a better starting geometry?
    output {
        coarse_geom = coarse_geom.expt
            .type = str
            .help = file to which the coarse geometry would be saved
        calibrated_geom = calibrated.expt
            .type = str
            .help = file to which the calibrated geometry would be saved
        pyfai_improvement = pyfai_improvement.png
            .type = str
            .help = file name for the pyfai calibration effect figure
        straight_lines = straight_lines.png
            .type = str
            .help = file name for the cake plot showing hopefully straight lines
        log = dials.powder_calibrate.log
           .type = path
    }
    """
)


class Point(NamedTuple):
    """
    Use the blessing of namedtuple for x y coordinates
    """

    fast: float
    slow: float


class ExptParams(NamedTuple):
    """
    Store the .expt parameters that will be used in calibration
    """

    input_file: str
    s0: Tuple[float, float, float]
    beam_on_detector: Point
    wavelength: float
    distance: float
    img_size: Tuple[int, int]
    pix_size: Tuple[float, float]
    image: np.ndarray


class Output(NamedTuple):
    """
    Store the names of the variety of .expt and .png output files
    """

    coarse_geom: str
    calibrated_geom: str
    pyfai_improv: str
    strt_lines: str


def _convert_units(
    val: Union[float, Point], unit_in: str, unit_out: str
) -> Union[float, Point]:
    """Keep units sanity for pyFAI <--> DIALS conversions
    Most parameters will be kept in SI units internally

    :param val: the value/values to be converted
    :param unit_in: Units of input value
    :param unit_out: Units wanted for the input value
    :return: value in new units
    """
    si = {"A": 1e-10, "nm": 1e-9, "micron": 1e-6, "mm": 1e-3, "cm": 1e-2, "m": 1}
    if isinstance(val, Point):
        return Point(*np.array(val) * si[unit_in] / si[unit_out])
    else:
        return val * si[unit_in] / si[unit_out]


def get_expt_params(expts: ExperimentList) -> ExptParams:
    """
    Read input.expt and return custom named tuple with the useful parameters
    """

    expt = expts[0]
    detector = expt.detector[0]
    beam = expt.beam

    s0 = beam.get_s0()
    beam_on_detector = detector.get_beam_centre(s0)

    distance = detector.get_distance()
    wavelength = beam.get_wavelength()
    pix_size = detector.get_pixel_size()

    # Detector size has dimensions flipped wrt data array
    img_size = detector.get_image_size()[::-1]
    image = np.array(expt.imageset.get_corrected_data(0)[0]).reshape(img_size)

    if hasattr(expt.imageset, "get_template"):
        file_name = expt.imageset.get_template()
    else:
        file_name = expt.imageset.data().get_master_path()

    expt_params = ExptParams(
        input_file=file_name,
        s0=s0,
        beam_on_detector=Point(*beam_on_detector),
        wavelength=wavelength,
        distance=distance,
        img_size=img_size,
        pix_size=pix_size,
        image=image,
    )

    return expt_params


class Detector(pfDetector):
    def __init__(self, expt_params: ExptParams):
        px, py = expt_params.pix_size
        det_shape = expt_params.img_size

        super().__init__(
            pixel1=_convert_units(px, "mm", "m"),
            pixel2=_convert_units(py, "mm", "m"),
            max_shape=det_shape,
        )


class Geometry(pfGeometry):
    """
    pyFAI uses the normal from the sample to the detector as the direction
    along which the detector distance and its intersection with the detector
    denotes the detector geometry. This intersection is labelled as "poni" and
    must be differentiated from the intersection of the beam with the detector.
    These two are coincident only in the case of flat (zero tilt) detector.

    Passing geometry back and from DIALS involves passing parameters through
    pyFAI.geometry.setFit2D which understands beam center as a geometry parameter
    and updates correspondingly the pyFAI.geometry and pyFAI.detector objects.
    Similarly, PyFAI.geometry.setFit2D does the reverse.
    """

    def __init__(self, expt: ExperimentList):
        self.expt = expt
        self.expt_params = get_expt_params(expt)

        super().__init__(
            detector=Detector(self.expt_params),
            wavelength=_convert_units(self.expt_params.wavelength, "A", "m"),
        )
        # convert beam parameters from mm to m
        self.beam_m = _convert_units(self.expt_params.beam_on_detector, "mm", "m")
        self.beam_px = self._beam_to_px()

        self.beam_distance = self.expt_params.distance

        # pyFAI calibration will need its poni parameters
        self._set_poni()

    def __repr__(self):
        return f"<Geometry({self.expt_params.input_file})>"

    def __str__(self):
        return (
            f"Geometry instance with following current parameters:\n"
            f"  Detector: \n"
            f"        pixel size = {_convert_units(self.detector.pixel1, 'm', 'mm')} mm x {_convert_units(self.detector.pixel2, 'm', 'mm')} mm \n"
            f"        image size = {self.detector.max_shape[0]} px x {self.detector.max_shape[1]} px\n"
            f"        distance along beam = {_convert_units(self.beam_distance, 'mm', 'm'):.2f} m\n"
            f"  Wavelength = {_convert_units(self.wavelength, 'm', 'A'):.4f} A\n"
            f"  Beam position on detector: (fast, slow) = ({self.beam_m.fast:.4f}, {self.beam_m.slow:.4f})m or"
            f" ({self.beam_px.fast:.2f}, {self.beam_px.slow:.2f})px"
        )

    def _set_poni(self):
        """
        Call fit2D to translate beam parameters to poni parameters
        """
        self.setFit2D(
            directDist=self.beam_distance,
            centerX=self.beam_px.fast,
            centerY=self.beam_px.slow,
        )

    def _beam_to_px(self):
        """
        Transform beam coordinates from meters to pixels
        """
        if not self.beam_m:
            exit("No beam information provided")
        return Point(
            fast=self.beam_m.fast / self.detector.pixel1,
            slow=self.beam_m.slow / self.detector.pixel2,
        )

    def _beam_to_m(self):
        """
        Transforms beam coordinates from pixels to meters
        """
        if not self.beam_px:
            exit("No beam information provided")
        return Point(
            fast=self.beam_px.fast * self.detector.pixel1,
            slow=self.beam_px.slow * self.detector.pixel2,
        )

    def update_beam_pos(
        self,
        beam_coords_px: Optional[Point] = None,
        beam_dist_mm: Optional[float] = None,
    ):
        """
        Update beam position

        :param beam_coords_px: new beam position in pixels
        :param beam_dist_mm: new detector distance along beam direction
        """

        self.beam_px = beam_coords_px
        self.beam_m = self._beam_to_m()

        if beam_dist_mm:
            self.beam_distance = beam_dist_mm

        self._set_poni()

    def update_from_ai(self, geom: pfGeometry):
        """
        Update geometry parameters from refined ai

        :param geom: pfGeometry object
        """
        self.set_param(geom.param)
        beam_params = self.getFit2D()

        self.beam_px = Point(fast=beam_params["centerX"], slow=beam_params["centerY"])

        self.beam_m = self._beam_to_m()
        self.beam_distance = beam_params["directDist"]

    def modify_geom_params(self) -> scope_extract:
        """
        Make a new set of phil parameters by modifying beam position, beam distance,
        and beam wavelength in the geometry phil
        """
        param = dials.command_line.modify_geometry.phil_scope.fetch().extract()
        param.geometry.detector.fast_slow_beam_centre = self.beam_px
        param.geometry.detector.distance = self.beam_distance
        param.geometry.beam.wavelength = _convert_units(self.wavelength, "m", "A")

        return param

    def save_to_expt(self, output: str | os.PathLike):
        """
        Update the geometry from start_geometry.expt and save to new output.
        Output is passed either as a path or str
        """
        new_params = self.modify_geom_params()
        new_geometry = geometry_update(experiments=self.expt, new_params=new_params)
        new_geometry.as_file(output)


class EyeballWidget:
    """
    Matplotlib widget to guestimate the beam center by comparing the theoretical
    and experimental standard.
    """

    def __init__(
        self,
        expt_params: ExptParams,
        start_geometry: Geometry,
        calibrant: pfCalibrant,
        coarse_geom: str,
    ):
        """Eyeball widget
        :param expt_params: experimental parameters
        :param start_geometry: initial geometry
        :param calibrant: the standard used for calibration
        :param coarse_geom: the file name to which the eyeballed geometry is to saved
        """
        self.input_model = expt_params.input_file
        self.image = expt_params.image
        self.geometry = start_geometry
        self.detector = start_geometry.detector
        self.calibrant = calibrant
        self.coarse_geom = coarse_geom

        self.fig, self.ax = self.set_up_figure()
        self.calibrant_image = self.calibrant_rings_image()

        self.beam_fast_slider = self._make_slider("fast")
        self.beam_slow_slider = self._make_slider("slow")
        self.distance_slider = self._make_slider("distance")

    def __repr__(self):
        return (
            f"EyeballWidget(image = {self.input_model},\n"
            f"              start_geometry = \n"
            f"{self.geometry},\n\n"
            f"              calibrant = {self.calibrant},\n"
            f"              coarse_geom = {self.coarse_geom})"
        )

    def __str__(self):
        calibrant = os.path.splitext(os.path.basename(self.calibrant.filename))[0]
        return f"Eyeballing Widget instance using {calibrant} Calibrant starting from Geometry: \n {self.geometry}"

    def set_up_figure(self):
        """
        Add title and label axes
        """
        # often the ED data will have negative intensities
        # ignore them for calibration purposes
        self.image[self.image <= 0] = 0

        fig, ax = plt.subplots()
        ax.imshow(np.arcsinh(self.image), origin="lower", cmap="inferno")

        ax.set_xlabel("fast position [pixels]")
        ax.set_ylabel("slow position [pixels]")
        fig.set_size_inches(12, 10)
        return fig, ax

    def _make_slider(self, label):
        slider = None

        if label == "fast":
            plt.subplots_adjust(bottom=0.25)
            # Make a horizontal slider to control the beam fast position.
            ax_beam_fast = plt.axes([0.25, 0.1, 0.65, 0.03])

            slider = Slider(
                ax=ax_beam_fast,
                label="Beam center fast [px]",
                valmin=self.detector.max_shape[0] * 0.25,
                valmax=self.detector.max_shape[0] * 0.75,
                valinit=self.geometry.beam_px.fast,
            )

        elif label == "slow":
            plt.subplots_adjust(left=0.25)
            # Make a vertical slider to control the beam slow position
            ax_beam_slow = plt.axes([0.1, 0.25, 0.0225, 0.63])

            slider = Slider(
                ax=ax_beam_slow,
                label="Beam center slow [px]",
                valmin=self.detector.max_shape[1] * 0.25,
                valmax=self.detector.max_shape[1] * 0.75,
                valinit=self.geometry.beam_px.slow,
                orientation="vertical",
            )

        elif label == "distance":
            plt.subplots_adjust(right=0.75)
            # Make another vertical slider to control the detector distance
            ax_beam_dist = plt.axes([0.9, 0.25, 0.0225, 0.63])

            slider = Slider(
                ax=ax_beam_dist,
                label="Detector beam distance [mm]",
                valmin=min(100.0, self.geometry.beam_distance),
                valmax=max(1000.0, self.geometry.beam_distance),
                valinit=self.geometry.beam_distance,
                orientation="vertical",
            )

        return slider

    def _update_with_slider(self, val):
        """
        Update calibrant geometry from slider values
        """
        self.geometry.update_beam_pos(
            beam_coords_px=Point(
                fast=self.beam_fast_slider.val, slow=self.beam_slow_slider.val
            ),
            beam_dist_mm=self.distance_slider.val,
        )

        # Update calibrant rings
        self.calibrant_image.set_array(self.calibrant_rings(self.geometry))

        # Redraw current figure
        self.fig.canvas.draw_idle()

    def _reset_button(self, event):
        """
        Reset calibrant rings image to starting position
        """
        self.beam_fast_slider.reset()
        self.beam_slow_slider.reset()
        self.distance_slider.reset()

    def _save_and_exit_button(self, event):
        """
        Save geometry from widget and save to file
        """
        self.geometry.save_to_expt(
            output=self.coarse_geom,
        )

        plt.close()

    def eyeball(self):
        """
        Update geometry such that beam position is now the center of the moved calibrant rings
        """
        # Register the update function with each slider
        self.beam_fast_slider.on_changed(self._update_with_slider)
        self.beam_slow_slider.on_changed(self._update_with_slider)
        self.distance_slider.on_changed(self._update_with_slider)

        # Register the reset function with reset button
        reset_ax = plt.axes([0.8, 0.026, 0.1, 0.04])
        button_res = Button(reset_ax, "Reset", hovercolor="0.975")
        button_res.on_clicked(self._reset_button)

        # Register the save and exit function with save button
        save_ax = plt.axes([0.5, 0.026, 0.23, 0.04])
        button_save = Button(save_ax, "Save beam and exit", hovercolor="0.975")
        button_save.on_clicked(self._save_and_exit_button)

        # Finally, show widget
        plt.show()

    def calibrant_rings(self, geometry: Optional[Geometry] = None) -> np.ndarray:
        """
        Make a masked array for the calibration rings. If a geometry parameters
        is given then use that geometry for the calibrant, otherwise use the
        class geometry.
        For smaller wavelengths (electrons) reduce the rings blurring
        """

        if self.calibrant.wavelength < 1e-1:  # electrons
            w = 1e-7
        else:
            w = 1e-3

        if geometry:
            cal_img = self.calibrant.fake_calibration_image(geometry, W=w)
        else:
            cal_img = self.calibrant.fake_calibration_image(self.geometry, W=w)

        # get rid of background
        cal_img_masked = ma.masked_where(cal_img == 0, cal_img)

        return cal_img_masked

    def calibrant_rings_image(self) -> plt.axes:
        cal_img_masked = self.calibrant_rings()

        rings_img = self.ax.imshow(
            np.arcsinh(cal_img_masked), alpha=0.2, cmap="inferno", origin="lower"
        )

        # Add the beam position
        self.ax.plot(*self.geometry.beam_px, "rx")

        return rings_img


class PowderCalibrator:
    available_calibrants = [
        "Al",
        "LaB6",
        "TiO2",
        "Pt",
        "Ni",
        "CuO",
        "quartz",
        "Si",
        "mock",
        "Si_SRM640e",
        "LaB6_SRM660a",
        "BBA",
        "cristobaltite",
        "Si_SRM640",
        "NaCl",
        "AgBh",
        "CrOx",
        "LaB6_SRM660c",
        "C14H30O",
        "Si_SRM640a",
        "Au",
        "alpha_Al2O3",
        "ZnO",
        "Si_SRM640d",
        "Cr2O3",
        "Si_SRM640c",
        "LaB6_SRM660b",
        "Si_SRM640b",
        "hydrocerussite",
        "CeO2",
    ]

    def __init__(
        self,
        expts: ExperimentList,
        standard: str,
        coarse_geometry: str = "coarse_geom.expt",
        calibrated_geometry: str = "calibrated.expt",
        pyfai_improvement: str = "pyfai_improvement.png",
        straight_lines: str = "straight_lines.png",
    ):
        """
        Perform geometry calibration using an electron powder standard. Because electron powder
        rings are more dense than X-ray ones, pyFAI struggles to automatically find the correct rings.
        The current method tries to ensure pyFAI has a good starting guess by asking the user to
        eyeball the beam position using a matplotlib widget which overlaps the theoretical calibrator rings
        over the measured ones. The widget part can be turned off setting "eyeball=False".

        :param expts: ExperimentList
                experiment list containing starting geometry
        :param standard: str
                calibrating standard name in periodic table name format.
                Call calibrator.show_calibrants to see available calibrants.
        :param coarse_geometry: str
                optional, if given saves the coarse geometry to this file
        :param calibrated_geometry: str
                optional, if given saves the calibrated geometry to this file
        :param pyfai_improvement: str
                optional, file name used for saving the before and after pyfai calibration
        :param straight_lines: str
                optional, file name used for saving the cake plot, if fir was successful
                these lines should be straight

        Examples
        --------
        1. Use the widget to generate a starting, coarse geometry after which the fine calibration
        is applied. Overwrite the default name of the coarse geometry output file. Do this from command line:

        .. code-block:: console
           dials.powder_calibrate poor_geom.expt standard="Al" coarse_geom="a_starting_geometry.expt"

        2. Starting from a coarse geometry file, perform fine calibration with pyFAI bypassing the
        widget tool. Do this using python API:

        .. code-block:: python
           from dials.command_line.powder_calibrate import PowderCalibrator
           al_calibrator = PowderCalibrator(expts="coarse_geom.expt", standard="Al", eyeball=False)
           al_calibrator.calibrate_with_widget()
           al_calibrator.refine_with_pyfai(num_rings=5,
                                           fix=("rot1", "rot2", "rot3", "wavelength"),
                                           Imin=10,
                                           pts_per_deg=1.0,
                                           plots=True)

        3. Do the same from command line::

        .. code-block:: console
           dials.powder_calibrate coarse_geom.expt standard="Al" eyeball=False
        """
        self.expt = expts
        self.expt_params = get_expt_params(expts)
        self.geometry = Geometry(self.expt)
        self.detector = self.geometry.detector

        # Check if calibrant name makes sense, else complain
        if standard in PowderCalibrator.available_calibrants:
            self.standard = standard
        else:
            PowderCalibrator.list_calibrants()
            exit(f"The standard name {standard} was not recognised")

        self.calibrant = pfCalibrant(standard)
        self.calibrant.wavelength = _convert_units(
            self.expt_params.wavelength, "A", "m"
        )
        self.output = Output(
            coarse_geom=coarse_geometry,
            calibrated_geom=calibrated_geometry,
            pyfai_improv=pyfai_improvement,
            strt_lines=straight_lines,
        )
        self.print_hints()

    def __repr__(self):
        return (
            f"<PowderCalibrator(expts = {self.expt_params.input_file},\n"
            f"                  standard = {self.standard},\n"
            f"                  output = {self.output})>"
        )

    def __str__(self):
        return (
            f"{self.standard} PowderCalibrator for {self.expt_params.input_file} \n"
            f"Current geometry: \n {self.geometry}"
        )

    def print_hints(self):
        # Tell me which geometry I'm starting from
        logger.info(
            f"Initial geometry from {self.expt_params.input_file}:\n"
            f"-----\n"
            f"{self.geometry}\n"
        )

        # Tell me what calibrant I am using
        logger.info(
            f"Starting calibration using {self.standard}:\n"
            f"-----\n"
            f"{self.calibrant}\n"
        )

    def show_pyfai_improvement(
        self,
        before_geometry: pfSingleGeometry,
        after_geometry: pfSingleGeometry,
        show: bool = True,
    ):
        fig, (ax1, ax2) = plt.subplots(1, 2)
        fig.set_size_inches(20, 8)
        pfjupyter.display(sg=before_geometry, label="Before pyFAI refine", ax=ax1)
        pfjupyter.display(sg=after_geometry, label="After pyFAI refine", ax=ax2)

        # Show only a quarter of the image
        custom_xlim = (
            self.geometry.detector.max_shape[0] * 0.5,
            self.geometry.detector.max_shape[0] * 0.75,
        )
        custom_ylim = (
            self.geometry.detector.max_shape[1] * 0.5,
            self.geometry.detector.max_shape[1] * 0.75,
        )
        plt.setp((ax1, ax2), xlim=custom_xlim, ylim=custom_ylim)

        if show:
            plt.show()
        else:
            fig.savefig(self.output.pyfai_improv, bbox_inches="tight")
            plt.close(fig)

    @staticmethod
    def list_calibrants():
        print(f"Available calibrants: {PowderCalibrator.available_calibrants}")

    def show_straight_lines(self, ai: pfGeometry, show: bool = True):
        # Show the cake plot as well
        int2 = ai.integrate2d_ng(
            data=self.expt_params.image,
            npt_rad=500,
            npt_azim=360,
            unit="q_nm^-1",
        )
        fig, ax = plt.subplots()
        fig.set_size_inches(10, 10)
        pfjupyter.plot2d(int2, calibrant=self.calibrant, ax=ax)
        plt.title("Are those lines straight?")

        if show:
            plt.show()
        else:
            fig.savefig(self.output.strt_lines, bbox_inches="tight")
            plt.close(fig)

    def calibrate_with_widget(self):
        # Tell me if I'm using the EyeballWidget and how to use
        logger.info(
            "Using the eyeballing widget.\n"
            "Drag sliders to roughly overlap rings and then save for further fitting. \n"
            "If the detector distance is very off start with that."
        )
        # First use user eyes for rough fit
        eyeballing_widget = EyeballWidget(
            expt_params=self.expt_params,
            start_geometry=self.geometry,
            calibrant=self.calibrant,
            coarse_geom=self.output.coarse_geom,
        )
        eyeballing_widget.eyeball()

    def refine_with_pyfai(
        self,
        num_rings: int = 5,
        fix: Tuple = ("rot1", "rot2", "rot3", "wavelength"),
        pts_per_deg: float = 1.0,
        Imin: float = 10.0,
        plots: bool = True,
    ):
        """
        Do the actual calibration

        :param num_rings: int
                number of Debye-Scherrer rings to fit starting from low resolution
        :param fix: tuple
                variables to keep fixed while fitting
                Note that, detector distance and wavelength are highly correlated,
                in order to calibrate both a higher number of rings must be used.
        :param pts_per_deg: float
                number of spots per radial degree to be searched for.
                If rings are sparse lower the number
        :param Imin: float
                minimum intensity for deciding something is a peak
        :param plots: bool
                show the fitting plots or keep it quiet
        """
        # Store starting geometry for future plotting
        starting_geom = pfSingleGeometry(
            label=self.standard + " calibrant in starting geom",
            image=self.expt_params.image,
            calibrant=self.calibrant,
            geometry=self.geometry,
        )

        # Pick spots around rings
        starting_geom.extract_cp(
            max_rings=num_rings, pts_per_deg=pts_per_deg, Imin=Imin
        )

        # Single geometry to refine
        # This obj is thread locked so not obvious how to deepcopy it,
        # make a fresh one instead
        gonio_geom = pfSingleGeometry(
            label=self.standard + " calibrant in calibrated geom",
            image=self.expt_params.image,
            calibrant=self.calibrant,
            geometry=self.geometry,
        )

        # Pick spots again
        gonio_geom.extract_cp(max_rings=num_rings, pts_per_deg=pts_per_deg, Imin=Imin)

        # Fit the data to the calibrant to refine geometry
        gonio_geom.geometry_refinement.refine2(fix=fix)

        # Generate plot showing the pyFAI refinement effect
        self.show_pyfai_improvement(
            before_geometry=starting_geom, after_geometry=gonio_geom, show=plots
        )

        # Update geometry and save to .expt
        ai = gonio_geom.get_ai()
        self.geometry.update_from_ai(ai)
        self.geometry.save_to_expt(output=self.output.calibrated_geom)

        logger.info(f"Geometry fitted by pyFAI:\n" f"-----\n" f"{self.geometry}\n")
        self.show_straight_lines(ai, show=plots)


@dials.util.show_mail_handle_errors()
def run(args: List[str] = None, phil: scope = phil_scope) -> None:
    """
    Check command-line input and do the sequence of
    function calls.

    :param args: The arguments supplied by the user (default: sys.argv[1:])
    :param phil: The PHIL scope definition (default: phil_scope, the master PHIL scope
    for this program).
    """
    usage = "$ dials.powder_calibrate imported.expt standard=standard_name [options]"

    parser = ArgumentParser(
        usage=usage, phil=phil, read_experiments=True, check_format=True, epilog=__doc__
    )

    parameters, options = parser.parse_args(args=args, show_diff_phil=False)

    experiments = flatten_experiments(parameters.input.experiments)

    # Check user is holding the tool right
    if not experiments or not parameters.standard:
        parser.print_help()
        exit()

    # Configure the logging.
    dials.util.log.config(options.verbose, logfile=parameters.output.log)

    # Log the dials version
    logger.info(dials_version())

    # Log the difference between the PHIL scope definition and the active PHIL scope,
    # which will include the parsed user inputs.
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    # Make a calibrator instance based on these parameters and then calibrate
    calibrator = PowderCalibrator(
        expts=experiments,
        standard=parameters.standard,
        coarse_geometry=parameters.output.coarse_geom,
        calibrated_geometry=parameters.output.calibrated_geom,
        pyfai_improvement=parameters.output.pyfai_improvement,
        straight_lines=parameters.output.straight_lines,
    )

    if parameters.eyeball:
        calibrator.calibrate_with_widget()

    calibrator.refine_with_pyfai(
        num_rings=5,
        fix=("rot1", "rot2", "rot3", "wavelength"),
        Imin=20,
        pts_per_deg=1.0,
        plots=True,
    )


if __name__ == "__main__":
    run()



"""
Test command line program dials.refine by running a job with saved data and
comparing with expected output.

This serves as a high level test that not only checks whether refinement works,
but also that the command line program is functioning and that the output models
have not changed format and so on.
"""


from __future__ import annotations

import os

import procrunner
import pytest
from annlib_ext import AnnAdaptor

from dxtbx.model.experiment_list import ExperimentListFactory

from dials.algorithms.refinement.engine import Journal
from dials.array_family import flex


def test1(dials_regression, tmp_path):
    # use the i04_weak_data for this test
    data_dir = os.path.join(dials_regression, "refinement_test_data", "i04_weak_data")
    experiments_path = os.path.join(data_dir, "experiments.json")
    pickle_path = os.path.join(data_dir, "indexed_strong.pickle")

    for pth in (experiments_path, pickle_path):
        assert os.path.exists(pth)

    # set some old defaults
    cmd = (
        "dials.refine",
        "close_to_spindle_cutoff=0.05",
        "reflections_per_degree=100",
        "outlier.separate_blocks=False",
        "scan_varying=False",
        "reflections.outlier.nproc=1",
        experiments_path,
        pickle_path,
    )
    result = procrunner.run(cmd, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    # load results
    reg_exp = ExperimentListFactory.from_json_file(
        os.path.join(data_dir, "regression_experiments.json"), check_format=False
    )[0]
    ref_exp = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )[0]

    # test refined models against expected
    assert reg_exp.crystal == ref_exp.crystal
    assert reg_exp.detector == ref_exp.detector
    assert reg_exp.beam == ref_exp.beam

    # test cell parameter esds
    assert ref_exp.crystal.get_cell_parameter_sd() == pytest.approx(
        (0.0009903, 0.0009903, 0.0021227, 0.0, 0.0, 0.0), abs=1e-6
    )
    assert ref_exp.crystal.get_cell_volume_sd() == pytest.approx(23.8063382, abs=1e-6)


def test2(dials_regression, tmp_path):
    """Run scan-varying refinement, comparing RMSD table with expected values.
    This test automates what was manually done periodically and recorded in
    dials_regression/refinement_test_data/centroid/README.txt"""

    # use the i04_weak_data for this test
    data_dir = os.path.join(dials_regression, "refinement_test_data", "centroid")
    experiments_path = os.path.join(data_dir, "experiments_XPARM_REGULARIZED.json")
    pickle_path = os.path.join(data_dir, "spot_all_xds.pickle")

    for pth in (experiments_path, pickle_path):
        assert os.path.exists(pth)

    # scan-static refinement first to get refined.expt as start point
    result = procrunner.run(
        (
            "dials.refine",
            experiments_path,
            pickle_path,
            "scan_varying=False",
            "reflections_per_degree=50",
            "outlier.algorithm=null",
            "close_to_spindle_cutoff=0.05",
        ),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    # NB Requesting corrgram.pdf and history.json exercises
    # https://github.com/dials/dials/issues/1923
    result = procrunner.run(
        (
            "dials.refine",
            "refined.expt",
            pickle_path,
            "scan_varying=true",
            "output.history=history.json",
            "correlation_plot.filename=corrgram.pdf",
            "reflections_per_degree=50",
            "outlier.algorithm=null",
            "close_to_spindle_cutoff=0.05",
            "crystal.orientation.smoother.interval_width_degrees=36.0",
            "crystal.unit_cell.smoother.interval_width_degrees=36.0",
            "set_scan_varying_errors=True",
            "reflections.outlier.nproc=1",
        ),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # load and check results
    history = Journal.from_json_file(tmp_path / "history.json")

    expected_rmsds = [
        (0.088488398, 0.114583571, 0.001460382),
        (0.080489334, 0.086406517, 0.001284069),
        (0.078835086, 0.086052630, 0.001195882),
        (0.077476911, 0.086194611, 0.001161143),
        (0.076755840, 0.086090630, 0.001157239),
        (0.076586376, 0.085939462, 0.001155641),
        (0.076603722, 0.085878953, 0.001155065),
        (0.076611382, 0.085862959, 0.001154863),
        (0.076608732, 0.085856935, 0.001154384),
        (0.076605731, 0.085852271, 0.001153858),
        (0.076604576, 0.085852318, 0.001153643),
        (0.076603981, 0.085854175, 0.001153594),
    ]
    for a, b in zip(history["rmsd"], expected_rmsds):
        assert a == pytest.approx(b, abs=1e-6)

    # check that the used_in_refinement flag got set correctly
    rt = flex.reflection_table.from_file(tmp_path / "refined.refl")
    uir = rt.get_flags(rt.flags.used_in_refinement)
    assert uir.count(True) == history["num_reflections"][-1]


def test3(dials_regression, tmp_path):
    """Strict check for scan-varying refinement using automated outlier rejection
    block width and interval width setting"""

    # use the i04_weak_data for this test
    data_dir = os.path.join(dials_regression, "refinement_test_data", "centroid")
    experiments_path = os.path.join(data_dir, "experiments_XPARM_REGULARIZED.json")
    pickle_path = os.path.join(data_dir, "spot_all_xds.pickle")

    for pth in (experiments_path, pickle_path):
        assert os.path.exists(pth)

    result = procrunner.run(
        (
            "dials.refine",
            experiments_path,
            pickle_path,
            "scan_varying=true",
            "max_iterations=5",
            "output.history=history.json",
            "crystal.orientation.smoother.interval_width_degrees=auto",
            "crystal.unit_cell.smoother.interval_width_degrees=auto",
            "reflections.outlier.nproc=1",
        ),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # load and check results
    history = Journal.from_json_file(tmp_path / "history.json")

    expected_rmsds = [
        [0.619507829, 0.351326044, 0.006955399],
        [0.174024575, 0.113486044, 0.004704006],
        [0.098351363, 0.084052519, 0.002660408],
        [0.069202909, 0.072796782, 0.001451734],
        [0.064305277, 0.071560831, 0.001165639],
        [0.062955462, 0.071315612, 0.001074453],
    ]
    for a, b in zip(history["rmsd"], expected_rmsds):
        assert a == pytest.approx(b, abs=1e-6)

    # check the refined unit cell
    ref_exp = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )[0]
    unit_cell = ref_exp.crystal.get_unit_cell().parameters()
    assert unit_cell == pytest.approx(
        [42.27482, 42.27482, 39.66893, 90.00000, 90.00000, 90.00000], abs=1e-3
    )

    refined_refl = flex.reflection_table.from_file(tmp_path / "refined.refl")
    # re-predict reflections using the refined experiments
    predicted = flex.reflection_table.from_predictions_multi([ref_exp])

    matched, reference, unmatched = predicted.match_with_reference(refined_refl)
    # assert most refined reflections are matched with predictions
    assert reference.size() > (0.997 * refined_refl.size())

    # second check with nearest neighbour matching that the predictions match up
    ann = AnnAdaptor(data=predicted["xyzcal.px"].as_double(), dim=3, k=1)
    ann.query(refined_refl["xyzcal.px"].as_double())
    assert (ann.distances < 0.5).count(True) > (0.998 * refined_refl.size())


@pytest.mark.parametrize("fix", ["cell", "orientation"])
def test_scan_varying_with_fixed_crystal(fix, dials_data, tmp_path):
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)
    refls = location / "reflections_1.pickle"
    expts = location / "experiments_1.json"

    result = procrunner.run(
        ("dials.refine", expts, refls, f"crystal.fix={fix}"), working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr


def test_scan_varying_missing_segments_multi_crystal(dials_data, tmp_path):
    # https://github.com/dials/dials/issues/1053
    location = dials_data("i19_1_pdteet_index", pathlib=True)
    refls = location / "indexed.refl"
    expts = location / "indexed.expt"

    # first remove some reflections to keep dials.refine sharp
    data = flex.reflection_table.from_file(refls)

    # prune only indexed reflections

    data = data.select(data.get_flags(data.flags.indexed))
    z = data["xyzobs.px.value"].parts()[2]

    # take overlapping subsets of reflections from two lattices which do not
    # fill the entire scan

    sel = ((data["id"] == 0) & (z > 200)) | ((data["id"] == 1) & (z < 650))
    trimmed = data.select(sel)

    trimmed_filename = tmp_path / "indexed_trim.refl"
    trimmed.as_file(trimmed_filename)

    result = procrunner.run(
        ("dials.refine", expts, trimmed_filename), working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr

    el = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )

    # verify scans start and finish correctly
    image_ranges = [e.scan.get_image_range() for e in el]

    assert image_ranges[0][1] == 850
    assert image_ranges[1][0] == 1


@pytest.mark.parametrize("gcb", ["None", "3000"])
def test_scan_varying_multi_scan_one_crystal(gcb, dials_data, tmp_path):
    # https://github.com/dials/dials/issues/994
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    refls = location / "indexed.refl"
    expts = location / "indexed.expt"

    # Set options for quick rather than careful refinement
    result = procrunner.run(
        (
            "dials.refine",
            expts,
            refls,
            "output.history=history.json",
            "outlier.algorithm=tukey",
            "max_iterations=3",
            "unit_cell.smoother.interval_width_degrees=56",
            "orientation.smoother.interval_width_degrees=56",
            "gradient_calculation_blocksize=" + gcb,
        ),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    el = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )

    # Crystal has been copied into each experiment for scan-varying refinement
    assert len(el.crystals()) == 4

    # load and check results
    history = Journal.from_json_file(tmp_path / "history.json")

    expected_rmsds = [
        (0.1401658782847504, 0.2225931584837884, 0.002349912655443814),
        (0.12060230585178289, 0.1585977879739876, 0.002114318828411418),
        (0.10970832317567975, 0.1348574975434352, 0.001955034565537597),
        (0.10373159352273859, 0.12827852889951505, 0.0017901404193256304),
    ]
    for a, b in zip(history["rmsd"], expected_rmsds):
        assert a == pytest.approx(b, abs=1e-6)



from __future__ import annotations

import random


class TestExact:
    def setup_class(self):
        self.size = (9, 10, 11)

    def test_constant2d_modeller(self):
        from dials.algorithms.background.simple import Constant2dModeller

        modeller = Constant2dModeller()
        eps = 1e-7
        for i in range(10):
            c, data, mask = self.generate_constant_background_2d(self.size, 0, 100)
            model = modeller.create(data, mask)
            assert len(model.params()) == 9
            for j in range(9):
                assert abs(model.params()[j] - c[j]) < eps

    def test_constant3d_modeller(self):
        from dials.algorithms.background.simple import Constant3dModeller

        modeller = Constant3dModeller()
        eps = 1e-7
        for i in range(10):
            c, data, mask = self.generate_constant_background_3d(self.size, 0, 100)
            model = modeller.create(data, mask)
            assert len(model.params()) == 1
            for j in range(1):
                assert abs(model.params()[j] - c) < eps

    def test_linear2d_modeller(self):
        from dials.algorithms.background.simple import Linear2dModeller

        modeller = Linear2dModeller()
        eps = 1e-7
        for i in range(10):
            p, data, mask = self.generate_linear_background_2d(self.size, 0, 100)
            model = modeller.create(data, mask)
            assert len(model.params()) == 3 * 9
            for j in range(9):
                for k in range(3):
                    assert abs(model.params()[k + j * 3] - p[j][k]) < eps

    def test_linear3d_modeller(self):
        from dials.algorithms.background.simple import Linear3dModeller

        modeller = Linear3dModeller()
        eps = 1e-7
        for i in range(10):
            p, data, mask = self.generate_linear_background_3d(self.size, 0, 100)
            model = modeller.create(data, mask)
            assert len(model.params()) == 4
            for j in range(4):
                assert abs(model.params()[j] - p[j]) < eps

    def generate_constant_background_2d(self, size, bmin, bmax):
        from scitbx.array_family import flex

        data = flex.double(flex.grid(size), 0)
        mask = flex.bool(flex.grid(size), True)
        slice_size = (1, size[1], size[2])
        cs = []
        for k in range(size[0]):
            c = random.uniform(bmin, bmax)
            data[k : k + 1, :, :] = flex.double(flex.grid(slice_size), c)
            cs.append(c)
        return cs, data, mask

    def generate_constant_background_3d(self, size, bmin, bmax):
        from scitbx.array_family import flex

        c = random.uniform(bmin, bmax)
        data = flex.double(flex.grid(size), c)
        mask = flex.bool(flex.grid(size), True)
        return c, data, mask

    def generate_linear_background_2d(self, size, bmin, bmax):
        from scitbx import matrix
        from scitbx.array_family import flex

        data = flex.double(flex.grid(size), 0)
        mask = flex.bool(flex.grid(size), True)
        params = []
        for k in range(size[0]):
            a00 = random.uniform(bmin, bmax)
            a01 = random.uniform(bmin, bmax)
            a10 = random.uniform(bmin, bmax)
            p00 = matrix.col((0.5, 0.5, a00))
            p01 = matrix.col((8.5, 0.5, a01))
            p10 = matrix.col((0.5, 8.5, a10))
            n = (p01 - p00).cross(p10 - p00)
            b = n[0]
            c = n[1]
            d = n[2]
            a = -(b * 0.5 + c * 0.5 + d * a00)
            a /= -d
            b /= -d
            c /= -d
            for j in range(size[1]):
                for i in range(size[2]):
                    data[k, j, i] = a + b * (i + 0.5) + c * (j + 0.5)
            eps = 1e-7
            assert abs(data[k, 0, 0] - a00) < eps
            assert abs(data[k, 8, 0] - a10) < eps
            assert abs(data[k, 0, 8] - a01) < eps
            params.append((a, b, c))
        return params, data, mask

    def generate_linear_background_3d(self, size, bmin, bmax):
        from scitbx import matrix
        from scitbx.array_family import flex

        data = flex.double(flex.grid(size), 0)
        mask = flex.bool(flex.grid(size), True)
        a000 = random.uniform(bmin, bmax)
        a001 = random.uniform(bmin, bmax)
        a010 = random.uniform(bmin, bmax)
        a100 = random.uniform(bmin, bmax)
        p000 = matrix.col((0.5, 0.5, 0.5, a000))
        p001 = matrix.col((8.5, 0.5, 0.5, a001))
        p010 = matrix.col((0.5, 8.5, 0.5, a010))
        p100 = matrix.col((0.5, 0.5, 8.5, a100))
        v1 = p001 - p000
        v2 = p010 - p000
        v3 = p100 - p000
        m1 = matrix.sqr((v1[1], v1[2], v1[3], v2[1], v2[2], v2[3], v3[1], v3[2], v3[3]))
        m2 = matrix.sqr((v1[0], v1[2], v1[3], v2[0], v2[2], v2[3], v3[0], v3[2], v3[3]))
        m3 = matrix.sqr((v1[0], v1[1], v1[3], v2[0], v2[1], v2[3], v3[0], v3[1], v3[3]))
        m4 = matrix.sqr((v1[0], v1[1], v1[2], v2[0], v2[1], v2[2], v3[0], v3[1], v3[2]))
        n = matrix.col(
            (m1.determinant(), -m2.determinant(), m3.determinant(), -m4.determinant())
        )
        b = n[0]
        c = n[1]
        d = n[2]
        e = n[3]
        a = -(b * 0.5 + c * 0.5 + d * 0.5 + e * a000)
        a /= -e
        b /= -e
        c /= -e
        d /= -e
        for k in range(size[0]):
            for j in range(size[1]):
                for i in range(size[2]):
                    data[k, j, i] = a + b * (i + 0.5) + c * (j + 0.5) + d * (k + 0.5)
        eps = 1e-7
        assert abs(data[0, 0, 0] - a000) < eps
        assert abs(data[8, 0, 0] - a100) < eps
        assert abs(data[0, 8, 0] - a010) < eps
        assert abs(data[0, 0, 8] - a001) < eps
        return (a, b, c, d), data, mask


class TestPoisson:
    def setup_class(self):
        self.size = (9, 9, 9)

    def test_constant2d_modeller(self):
        from dials.algorithms.background.simple import Constant2dModeller
        from dials.array_family import flex

        modeller = Constant2dModeller()
        ma = 10
        sboxes, masks = self.generate_background(self.size, 1000, ma, 0, 0, 0)
        a = []
        v = []
        for i in range(1000):
            model = modeller.create(sboxes[i], masks[i])
            assert len(model.params()) == 9
            assert len(model.variances()) == 9
            a.extend(list(model.params()))
            v.extend(list(model.variances()))

        # Compute Z for each parameter
        z = (flex.double(a) - ma) / flex.sqrt(flex.double(v))

        # Check it looks standard normal
        self.assert_std_norm(z)

    def test_constant3d_modeller(self):
        from dials.algorithms.background.simple import Constant3dModeller
        from dials.array_family import flex

        modeller = Constant3dModeller()

        ma = 10
        sboxes, masks = self.generate_background(self.size, 1000, ma, 0, 0, 0)
        a = []
        v = []
        for i in range(1000):
            model = modeller.create(sboxes[i], masks[i])
            assert len(model.params()) == 1
            assert len(model.variances()) == 1
            a.append(model.params()[0])
            v.append(model.variances()[0])

        # Compute Z for each parameter
        z = (flex.double(a) - ma) / flex.sqrt(flex.double(v))

        # Check it looks standard normal
        self.assert_std_norm(z)

    def test_linear2d_modeller(self):
        from dials.algorithms.background.simple import Linear2dModeller
        from dials.array_family import flex

        modeller = Linear2dModeller()

        # Generate shoeboxes
        ma = 100
        mb = 1
        mc = 2
        sboxes, masks = self.generate_background(self.size, 1000, ma, mb, mc, 0)

        pa = []
        pv = []
        for i in range(1000):
            model = modeller.create(sboxes[i], masks[i])
            assert len(model.params()) == 9 * 3
            assert len(model.variances()) == 9 * 3
            p = model.params()
            v = model.variances()
            for j in range(9):
                pa.append(tuple(p[3 * j : 3 * (j + 1)]))
                pv.append(tuple(v[3 * j : 3 * (j + 1)]))
        a, b, c = zip(*pa)
        va, vb, vc = zip(*pv)

        # Compute Z for each parameter
        za = (flex.double(a) - ma) / flex.sqrt(flex.double(va))
        zb = (flex.double(b) - mb) / flex.sqrt(flex.double(vb))
        zc = (flex.double(c) - mc) / flex.sqrt(flex.double(vc))

        # Check it looks standard normal
        self.assert_std_norm(za)
        self.assert_std_norm(zb)
        self.assert_std_norm(zc)

    def test_linear3d_modeller(self):
        from dials.algorithms.background.simple import Linear3dModeller
        from dials.array_family import flex

        modeller = Linear3dModeller()

        # Generate shoeboxes
        ma = 100
        mb = 1
        mc = 2
        md = 3
        sboxes, masks = self.generate_background(self.size, 1000, ma, mb, mc, md)

        # Compute model
        pa = []
        pv = []
        for i in range(1000):
            model = modeller.create(sboxes[i], masks[i])
            assert len(model.params()) == 4
            assert len(model.variances()) == 4
            p = model.params()
            v = model.variances()
            pa.append(tuple(p))
            pv.append(tuple(v))
        a, b, c, d = zip(*pa)
        va, vb, vc, vd = zip(*pv)

        # Compute Z for each parameter
        za = (flex.double(a) - ma) / flex.sqrt(flex.double(va))
        zb = (flex.double(b) - mb) / flex.sqrt(flex.double(vb))
        zc = (flex.double(c) - mc) / flex.sqrt(flex.double(vc))
        zd = (flex.double(d) - md) / flex.sqrt(flex.double(vd))

        # Check it looks standard normal
        self.assert_std_norm(za)
        self.assert_std_norm(zb)
        self.assert_std_norm(zc)
        self.assert_std_norm(zd)

    def assert_std_norm(self, z):
        from dials.array_family import flex

        mv = flex.mean_and_variance(z)
        m = mv.mean()
        s = mv.unweighted_sample_standard_deviation()
        # Mean-test failure rate:
        # 5*P(abs(X/1000) > 0.1336) where X is normally distributed
        #   with mean 0 and standard deviation sqrt(1000)
        if (abs(m) > 0.1336) or (abs(s - 1.0) > 0.1):
            # from matplotlib import pylab
            # pylab.hist(list(z), 100)
            # pylab.show()
            raise Exception(f"Mean {m:f} (abs. value <0.1336), Sdev {s:f} (>0.9, <1.1)")

    def generate_background(self, size, N, A, B, C, D):
        from dials.algorithms.simulation.generate_test_reflections import (
            random_background_plane2,
        )
        from dials.array_family import flex
        from dials.util.command_line import ProgressBar

        sboxes = []
        masks = []
        progress = ProgressBar(title="Generating Background")
        for i in range(N):
            mask = flex.bool(flex.grid(size), True)
            sbox = flex.double(flex.grid(size), 0)
            random_background_plane2(sbox, A, B, C, D)
            sboxes.append(sbox)
            masks.append(mask)
            progress.update(100.0 * i / N)
        progress.finished("Generated Background")
        return sboxes, masks



"""Calculate a global background model and diagnostic images from integrated
experiments. This background model may improve subsequent integration in some
cases, such as in the presence of ice rings. The method is described in the
publication https://doi.org/10.1107/S2052252517010259.

Usage:
    dials.model_background integrated.expt
    dials.integrate integrated.expt refined.refl background.algorithm=gmodel gmodel.robust.algorithm=True gmodel.model=background.pickle
"""


from __future__ import annotations

import logging
import pickle
import sys

from libtbx.phil import parse

import dials.util
import dials.util.log

logger = logging.getLogger("dials.command_line.model_background")

# Set the phil scope
phil_scope = parse(
    """

  output {
    model = 'background.pickle'
      .type = str
      .help = "The output filename"

    log = 'dials.model_background.log'
      .type = str
      .help = "The log filename"

    mean_image_prefix = 'mean'
      .type = str
      .help = "The mean background image"

    variance_image_prefix = 'variance'
      .type = str
      .help = "The variance background image"

    dispersion_image_prefix = 'dispersion'
      .type = str
      .help = "The dispersion background image"

    mask_image_prefix = 'mask'
      .type = str
      .help = "The mask background image"

    min_image_prefix = 'min'
      .type = str
      .help = "The min background image"

    max_image_prefix = 'max'
      .type = str
      .help = "The max background image"

    model_image_prefix = 'model'
      .type = str
      .help = "The model background image"

    polar_model_image_prefix = 'polar'
      .type = str
      .help = "The polar model background image"
  }

  modeller {

    min_images = 10
      .type = int(value_min=1)
      .help = "The minimum number of images per pixel"

    filter_type = *median mean
      .type = choice
      .help = "The filter to use on the polar transformed image"

    kernel_size = 50
      .type = int(value_min=0)
      .help = "The kernel size for the median filter"

    niter = 100
      .type = int(value_min=1)
      .help = "The number of iterations for filling holes"

    image_type = min *mean
      .type = choice
      .help = "Which image to use"

  }

  include scope dials.algorithms.integration.integrator.phil_scope
  include scope dials.algorithms.spot_prediction.reflection_predictor.phil_scope
""",
    process_includes=True,
)


class ImageGenerator:
    """
    Generate diagnostic images
    """

    def __init__(self, model):
        """
        Init the model
        """
        import matplotlib

        matplotlib.use("Agg")
        self.model = model

    def _save_plot(
        self, name, filename, extractor_fn, bounded=True, colorbar=True, vmax=None
    ):
        """
        Save the image
        """
        from matplotlib import pylab

        for i, model in enumerate(self.model):
            image = extractor_fn(model)
            pylab.figure(figsize=(6, 4))
            if bounded and vmax is None:
                boundaries = {
                    "vmin": 0,
                    "vmax": sorted(image)[int(0.99 * len(image))],
                }
            elif bounded:
                boundaries = {"vmin": 0, "vmax": vmax}
            else:
                boundaries = {}
            pylab.imshow(image.as_numpy_array(), interpolation="none", **boundaries)
            ax1 = pylab.gca()
            ax1.get_xaxis().set_visible(False)
            ax1.get_yaxis().set_visible(False)
            if colorbar:
                cb = pylab.colorbar()
                cb.ax.tick_params(labelsize=8)
            logger.info(
                "Saving %s image for panel %d to %s_%d.png", name, i, filename, i
            )
            pylab.savefig("%s_%d.png" % (filename, i), dpi=600, bbox_inches="tight")

    def save_min(self, filename):
        """
        Save the min image
        """
        self._save_plot("min", filename, lambda m: m.min_image)

    def save_max(self, filename):
        """
        Save the max image
        """
        self._save_plot("max", filename, lambda m: m.max_image)

    def save_mean(self, filename):
        """
        Save the mean image
        """
        self._save_plot("mean", filename, lambda m: m.mean)

    def save_variance(self, filename):
        """
        Save the variance image
        """
        self._save_plot("variance", filename, lambda m: m.variance)

    def save_dispersion(self, filename):
        """
        Save the dispersion image
        """
        self._save_plot("dispersion", filename, lambda m: m.dispersion, vmax=2)

    def save_mask(self, filename):
        """
        Save the dispersion image
        """
        self._save_plot(
            "mask", filename, lambda m: m.mask, bounded=False, colorbar=False
        )

    def save_model(self, filename):
        """
        Save the model image
        """
        self._save_plot("model", filename, lambda m: m.model)

    def save_polar_model(self, filename):
        """
        Save the polar model image
        """
        self._save_plot("polar model", filename, lambda m: m.polar_model, bounded=False)


class Script:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        from dials.util.options import ArgumentParser

        usage = "dials.model_background [options] [param.phil] models.expt"

        # Initialise the base class
        self.parser = ArgumentParser(
            usage=usage, phil=phil_scope, epilog=__doc__, read_experiments=True
        )

    def run(self, args=None):
        """Execute the script."""
        from dials.algorithms.background.modeller import BackgroundModeller
        from dials.array_family import flex
        from dials.util.command_line import heading
        from dials.util.options import flatten_experiments

        # Parse the command line
        params, options = self.parser.parse_args(args, show_diff_phil=False)

        # Configure the logging
        dials.util.log.config(verbosity=options.verbose, logfile=params.output.log)

        if params.integration.mp.nproc != 1 or params.integration.mp.njobs != 1:
            # https://github.com/dials/dials/issues/1083
            logger.warning(
                "Multiprocessing is currently disabled. " "Setting nproc = njobs = 1"
            )
            params.integration.mp.nproc = 1
            params.integration.mp.njobs = 1

        from dials.util.version import dials_version

        logger.info(dials_version())

        # Log the diff phil
        diff_phil = self.parser.diff_phil.as_str()
        if diff_phil != "":
            logger.info("The following parameters have been modified:\n")
            logger.info(diff_phil)

        # Ensure we have a data block
        experiments = flatten_experiments(params.input.experiments)
        if len(experiments) == 0:
            self.parser.print_help()
            return

        if any((e.profile is None for e in experiments)):
            sys.exit(
                "Experiments must contain a profile model (for example, after integration)"
            )

        # Only handle a single imageset at once
        imagesets = {expr.imageset for expr in experiments}
        if len(imagesets) != 1:
            sys.exit("Can only process a single imageset at a time")

        # Predict the reflections
        logger.info("")
        logger.info("=" * 80)
        logger.info("")
        logger.info(heading("Predicting reflections"))
        logger.info("")
        predicted = flex.reflection_table.from_predictions_multi(
            experiments,
            dmin=params.prediction.d_min,
            dmax=params.prediction.d_max,
            margin=params.prediction.margin,
            force_static=params.prediction.force_static,
        )

        # Create the modeller
        modeller = BackgroundModeller(experiments, predicted, params)
        model = modeller.compute()

        # Save the background model
        logger.info("Saving background model to %s", params.output.model)
        from dials.algorithms.background.gmodel import StaticBackgroundModel

        static_model = StaticBackgroundModel()
        for m in model:
            static_model.add(m.model)
        with open(params.output.model, "wb") as outfile:
            pickle.dump(static_model, outfile, protocol=pickle.HIGHEST_PROTOCOL)

        # Output some diagnostic images
        image_generator = ImageGenerator(model)
        image_generator.save_mean(params.output.mean_image_prefix)
        image_generator.save_variance(params.output.variance_image_prefix)
        image_generator.save_dispersion(params.output.dispersion_image_prefix)
        image_generator.save_mask(params.output.mask_image_prefix)
        image_generator.save_min(params.output.min_image_prefix)
        image_generator.save_max(params.output.max_image_prefix)
        image_generator.save_model(params.output.model_image_prefix)
        # image_generator.save_polar_model(params.output.polar_model_image_prefix)


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

import pickle

__all__ = ["reference", "reflections"]


def reflections(infile):
    """
    Load the given reflection file.

    :params infile: The input filename or file object
    :returns: The reflection list
    """

    # If the input is a string then open and read from that file
    if isinstance(infile, str):
        with open(infile, "rb") as infile:
            return pickle.load(infile)

    # Otherwise assume the input is a file and read from it
    else:
        return pickle.load(infile)


def reference(infile):
    """
    Load the given reference profile file.

    :params infile: The input filename or file object
    :returns: The reference list
    """

    # If the input is a string then open and read from that file
    if isinstance(infile, str):
        with open(infile, "rb") as infile:
            return pickle.load(infile)

    # Otherwise assume the input is a file and read from it
    else:
        return pickle.load(infile)



from __future__ import annotations

import logging
import math

from libtbx import phil
from rstbx.array_family import (
    flex,  # required to load scitbx::af::shared<rstbx::Direction> to_python converter
)
from rstbx.dps_core import SimpleSamplerTool
from scitbx import matrix

from dials.algorithms.indexing import DialsIndexError

from .strategy import Strategy
from .utils import group_vectors

logger = logging.getLogger(__name__)


real_space_grid_search_phil_str = """\
characteristic_grid = 0.02
    .type = float(value_min=0)
max_vectors = 30
    .help = "The maximum number of unique vectors to find in the grid search."
    .type = int(value_min=3)
"""


class RealSpaceGridSearch(Strategy):
    """
    Basis vector search using a real space grid search.

    Search strategy to index found spots based on known unit cell parameters. It is
    often useful for difficult cases of narrow-wedge rotation data or stills data,
    especially where there is diffraction from multiple crystals.

    A set of dimensionless radial unit vectors, typically ~7000 in total, is chosen
    so that they are roughly evenly spaced in solid angle over a hemisphere. For each
    direction, each of the three known unit cell vectors is aligned with the unit
    vector and is scored according to how well it accords with the periodicity in
    that direction of the reconstructed reciprocal space positions of the observed
    spot centroids. Examining the highest-scoring combinations, any basis vectors in
    orientations that are nearly collinear with a shorter basis vector are
    eliminated. The highest-scoring remaining combinations are selected as the basis
    of the direct lattice.

    See:
        Gildea, R. J., Waterman, D. G., Parkhurst, J. M., Axford, D., Sutton, G., Stuart, D. I., Sauter, N. K., Evans, G. & Winter, G. (2014). Acta Cryst. D70, 2652-2666.
    """

    phil_help = (
        "Index the found spots by testing a known unit cell in various orientations "
        "until the best match is found. This strategy is often useful for difficult "
        "cases of narrow-wedge rotation data or stills data, especially where there "
        "is diffraction from multiple crystals."
    )

    phil_scope = phil.parse(real_space_grid_search_phil_str)

    def __init__(self, max_cell, target_unit_cell, params=None, *args, **kwargs):
        """Construct a real_space_grid_search object.

        Args:
            max_cell (float): An estimate of the maximum cell dimension of the primitive
                cell.
            target_unit_cell (cctbx.uctbx.unit_cell): The target unit cell.
        """
        super().__init__(max_cell, params=params, *args, **kwargs)
        if target_unit_cell is None:
            raise DialsIndexError(
                "Target unit cell must be provided for real_space_grid_search"
            )
        self._target_unit_cell = target_unit_cell

    @property
    def search_directions(self):
        """Generator of the search directions (i.e. vectors with length 1)."""
        SST = SimpleSamplerTool(self._params.characteristic_grid)
        SST.construct_hemisphere_grid(SST.incr)
        for direction in SST.angles:
            yield matrix.col(direction.dvec)

    @property
    def search_vectors(self):
        """Generator of the search vectors.

        The lengths of the vectors correspond to the target unit cell dimensions.
        """
        unique_cell_dimensions = set(self._target_unit_cell.parameters()[:3])
        for i, direction in enumerate(self.search_directions):
            for l in unique_cell_dimensions:
                yield direction * l

    @staticmethod
    def compute_functional(vector, reciprocal_lattice_vectors):
        """Compute the functional for a single direction vector.

        Args:
            vector (tuple): The vector at which to compute the functional.
            reciprocal_lattice_vectors (scitbx.array_family.flex.vec3_double):
                The list of reciprocal lattice vectors.

        Returns:
            The functional for the given vector.
        """
        two_pi_S_dot_v = 2 * math.pi * reciprocal_lattice_vectors.dot(vector)
        return flex.sum(flex.cos(two_pi_S_dot_v))

    def score_vectors(self, reciprocal_lattice_vectors):
        """Compute the functional for the given directions.

        Args:
            directions: An iterable of the search directions.
            reciprocal_lattice_vectors (scitbx.array_family.flex.vec3_double):
                The list of reciprocal lattice vectors.
        Returns:
            A tuple containing the list of search vectors and their scores.
        """
        vectors = flex.vec3_double()
        scores = flex.double()
        for i, v in enumerate(self.search_vectors):
            f = self.compute_functional(v.elems, reciprocal_lattice_vectors)
            vectors.append(v.elems)
            scores.append(f)
        return vectors, scores

    def find_basis_vectors(self, reciprocal_lattice_vectors):
        """Find a list of likely basis vectors.

        Args:
            reciprocal_lattice_vectors (scitbx.array_family.flex.vec3_double):
                The list of reciprocal lattice vectors to search for periodicity.
        """
        used_in_indexing = flex.bool(reciprocal_lattice_vectors.size(), True)
        logger.info("Indexing from %i reflections", used_in_indexing.count(True))

        vectors, weights = self.score_vectors(reciprocal_lattice_vectors)

        perm = flex.sort_permutation(weights, reverse=True)
        vectors = vectors.select(perm)
        weights = weights.select(perm)

        groups = group_vectors(vectors, weights, max_groups=self._params.max_vectors)
        unique_vectors = []
        unique_weights = []
        for g in groups:
            idx = flex.max_index(flex.double(g.weights))
            unique_vectors.append(g.vectors[idx])
            unique_weights.append(g.weights[idx])

        logger.info("Number of unique vectors: %i", len(unique_vectors))

        for v, w in zip(unique_vectors, unique_weights):
            logger.debug("%s %s %s", w, v.length(), str(v.elems))

        return unique_vectors, used_in_indexing



from __future__ import annotations

import pytest

from cctbx import crystal, miller, sgtbx

from dials.algorithms.symmetry.cosym._generate_test_data import generate_intensities
from dials.algorithms.symmetry.laue_group import LaueGroupAnalysis


def generate_fake_intensities(crystal_symmetry):
    intensities = (
        generate_intensities(crystal_symmetry, d_min=1.0)
        .generate_bijvoet_mates()
        .set_observation_type_xray_intensity()
    )
    intensities = intensities.expand_to_p1()
    # needed to give vaguely sensible E_cc_true values
    intensities = intensities.customized_copy(sigmas=intensities.data() / 50)
    intensities.set_info(miller.array_info(source="fake", source_type="mtz"))
    return intensities


@pytest.mark.parametrize("space_group", ["P2", "P3", "P6", "R3:h", "I23"][:])
def test_determine_space_group(space_group):
    sgi = sgtbx.space_group_info(symbol=space_group)
    sg = sgi.group()
    cs = sgi.any_compatible_crystal_symmetry(volume=10000)
    cs = cs.best_cell()
    cs = cs.minimum_cell()

    intensities = generate_fake_intensities(cs)
    result = LaueGroupAnalysis([intensities], normalisation=None)
    print(result)
    assert (
        result.best_solution.subgroup["best_subsym"].space_group()
        == sg.build_derived_patterson_group()
    )
    assert result.best_solution.likelihood > 0.8
    for score in result.subgroup_scores[1:]:
        assert score.likelihood < 0.1


def test_determine_space_group_best_monoclinic_beta():
    cs = crystal.symmetry(
        unit_cell=(44.66208171, 53.12629403, 62.53397661, 64.86329707, 78.27343894, 90),
        space_group_symbol="C 1 2/m 1 (z,x+y,-2*x)",
    )
    cs = cs.best_cell().primitive_setting()
    intensities = generate_fake_intensities(cs)
    result = LaueGroupAnalysis([intensities], normalisation=None)
    print(result)
    assert result.best_solution.subgroup["best_subsym"].is_similar_symmetry(
        crystal.symmetry(
            unit_cell=(44.66208171, 53.12629403, 111.9989451, 90, 99.89337396, 90),
            space_group_symbol="I 1 2/m 1",
        )
    )
    d = result.as_dict()
    assert cs.change_basis(
        sgtbx.change_of_basis_op(d["subgroup_scores"][0]["cb_op"])
    ).is_similar_symmetry(result.best_solution.subgroup["best_subsym"])
    result = LaueGroupAnalysis(
        [intensities], best_monoclinic_beta=False, normalisation=None
    )
    print(result)
    assert result.best_solution.subgroup["best_subsym"].is_similar_symmetry(
        crystal.symmetry(
            unit_cell=(113.2236274, 53.12629403, 44.66208171, 90, 102.9736126, 90),
            space_group_symbol="C 1 2/m 1",
        )
    )
    d = result.as_dict()
    assert cs.change_basis(
        sgtbx.change_of_basis_op(d["subgroup_scores"][0]["cb_op"])
    ).is_similar_symmetry(result.best_solution.subgroup["best_subsym"])



from __future__ import annotations

import pytest

from scitbx import matrix

from dials.algorithms.indexing.basis_vector_search.utils import (
    group_vectors,
    is_approximate_integer_multiple,
    vector_group,
)


def test_is_approximate_integer_multiple():
    v1 = matrix.col((34, 50, 20))
    v2 = 2.1 * v1.rotate_around_origin(matrix.col((1, 0, 0)), 3, deg=True)
    assert is_approximate_integer_multiple(v1, v2)
    assert not is_approximate_integer_multiple(v1, v2, relative_length_tolerance=0.01)
    assert not is_approximate_integer_multiple(v1, v2, angular_tolerance=2)


def test_vector_group():
    group = vector_group()
    group.append(matrix.col((12, 13, 14)), weight=10)
    group.append(matrix.col((12.1, 13.1, 14.1)), weight=20)
    group.append(matrix.col((12.2, 13.2, 14.2)), weight=30)
    assert group.mean.elems == (12.1, 13.1, 14.1)
    assert group.weights == [10, 20, 30]


def test_group_vectors():
    vectors = [
        matrix.col(v)
        for v in (
            (12, 13, 14),
            (12.1, 13.1, 14.1),
            (12.2, 13.2, 14.2),
            (14, 15, 16),
            (14.1, 15.1, 16.1),
            (14.2, 15.2, 16.2),
        )
    ]
    weights = [1, 2, 3, 4, 5, 6]
    groups = group_vectors(vectors, weights)
    assert len(groups) == 2
    groups[0].mean.elems == pytest.approx((12.1, 13.1, 14.1))
    groups[0].weights == [1, 2, 3]
    groups[1].mean.elems == pytest.approx((14.1, 15.1, 16.1))
    groups[1].weights == [4, 5, 6]

    # test max_groups parameter
    groups = group_vectors(vectors, weights, max_groups=1)
    assert len(groups) == 1
    groups[0].mean.elems == pytest.approx((12.1, 13.1, 14.1))
    groups[0].weights == [1, 2, 3]



from __future__ import annotations


class BackgroundAlgorithm:
    """Class to do background subtraction."""

    def __init__(self, experiments, outlier="nsigma", model="constant3d", **kwargs):
        """
        Initialise the algorithm.

        :param experiments: The list of experiments
        :param outlier: The outlier rejection algorithm
        :param model: The background model algorithm
        """
        from dials.algorithms.background.simple import (
            Constant2dModeller,
            Constant3dModeller,
            Creator,
            Linear2dModeller,
            Linear3dModeller,
            MosflmOutlierRejector,
            NormalOutlierRejector,
            NSigmaOutlierRejector,
            TruncatedOutlierRejector,
            TukeyOutlierRejector,
        )

        def select_modeller():
            if model == "constant2d":
                return Constant2dModeller()
            elif model == "constant3d":
                return Constant3dModeller()
            elif model == "linear2d":
                return Linear2dModeller()
            elif model == "linear3d":
                return Linear3dModeller()
            raise RuntimeError(f"Unexpected background model: {model}")

        def select_rejector():
            if outlier == "null":
                return None
            elif outlier == "truncated":
                return TruncatedOutlierRejector(
                    kwargs.get("lower", 0.01), kwargs.get("upper", 0.01)
                )
            elif outlier == "nsigma":
                return NSigmaOutlierRejector(
                    kwargs.get("lower", 3), kwargs.get("upper", 3)
                )
            elif outlier == "normal":
                return NormalOutlierRejector(kwargs.get("min_pixels", 10))
            elif outlier == "plane":
                return MosflmOutlierRejector(
                    kwargs.get("fraction", 1.0), kwargs.get("n_sigma", 4.0)
                )
            elif outlier == "tukey":
                return TukeyOutlierRejector(
                    kwargs.get("lower", 1.5), kwargs.get("upper", 1.5)
                )
            raise RuntimeError(f"Unexpected outlier rejector: {outlier}")

        # Get the minimum number of pixels
        min_pixels = kwargs.get("min_pixels", 10)

        modeller = select_modeller()
        rejector = select_rejector()
        self._creator = Creator(modeller, rejector, min_pixels=min_pixels)

    def compute_background(self, reflections, image_volume=None):
        """
        Compute the background.

        :param reflections: The list of reflections
        """
        from dials.array_family import flex

        # Do the background subtraction
        if image_volume is None:
            reflections["background.mse"] = flex.double(len(reflections))
            reflections["background.dispersion"] = flex.double(len(reflections))
            success = self._creator(
                reflections["shoebox"],
                reflections["background.mse"],
                reflections["background.dispersion"],
            )
            reflections["background.mean"] = reflections["shoebox"].mean_background()
        else:
            success = self._creator(reflections, image_volume)
        reflections.set_flags(~success, reflections.flags.dont_integrate)
        return success


class SimpleBackgroundCalculatorFactory:
    """Class to do background subtraction."""

    @staticmethod
    def create(experiments, outlier="nsigma", model="constant3d", **kwargs):
        """
        Initialise the algorithm.

        :param experiments: The list of experiments
        :param outlier: The outlier rejection algorithm
        :param model: The background model algorithm
        """
        from dials.algorithms.background.simple import (
            Constant2dModeller,
            Constant3dModeller,
            Linear2dModeller,
            Linear3dModeller,
            MosflmOutlierRejector,
            NormalOutlierRejector,
            NSigmaOutlierRejector,
            TruncatedOutlierRejector,
            TukeyOutlierRejector,
        )
        from dials.algorithms.integration.parallel_integrator import (
            SimpleBackgroundCalculator,
        )

        def select_modeller():
            if model == "constant2d":
                return Constant2dModeller()
            elif model == "constant3d":
                return Constant3dModeller()
            elif model == "linear2d":
                return Linear2dModeller()
            elif model == "linear3d":
                return Linear3dModeller()
            raise RuntimeError(f"Unexpected background model: {model}")

        def select_rejector():
            if outlier == "null":
                return None
            elif outlier == "truncated":
                return TruncatedOutlierRejector(
                    kwargs.get("lower", 0.01), kwargs.get("upper", 0.01)
                )
            elif outlier == "nsigma":
                return NSigmaOutlierRejector(
                    kwargs.get("lower", 3), kwargs.get("upper", 3)
                )
            elif outlier == "normal":
                return NormalOutlierRejector(kwargs.get("min_pixels", 10))
            elif outlier == "plane":
                return MosflmOutlierRejector(
                    kwargs.get("fraction", 1.0), kwargs.get("n_sigma", 4.0)
                )
            elif outlier == "tukey":
                return TukeyOutlierRejector(
                    kwargs.get("lower", 1.5), kwargs.get("upper", 1.5)
                )
            raise RuntimeError(f"Unexpected outlier rejector: {outlier}")

        # Get the minimum number of pixels
        min_pixels = kwargs.get("min_pixels", 10)

        modeller = select_modeller()
        rejector = select_rejector()
        return SimpleBackgroundCalculator(modeller, rejector, min_pixels)



from __future__ import annotations

import random

import pytest

import scitbx.matrix
from cctbx import sgtbx
from cctbx.sgtbx import bravais_types

from dials.algorithms.indexing.max_cell import find_max_cell
from dials.array_family import flex


@pytest.fixture(params=bravais_types.acentric)
def setup(request):
    space_group_symbol = request.param
    sgi = sgtbx.space_group_info(space_group_symbol)
    cs = sgi.any_compatible_crystal_symmetry(volume=random.randint(10_000, 1_000_000))
    ms = cs.build_miller_set(anomalous_flag=True, d_min=3).expand_to_p1()

    # the reciprocal matrix
    B = scitbx.matrix.sqr(cs.unit_cell().fractionalization_matrix()).transpose()

    # randomly select 25% of reflections
    ms = ms.select(flex.random_permutation(ms.size())[: int(0.25 * ms.size())])

    refl = flex.reflection_table()
    refl["rlp"] = B.elems * ms.indices().as_vec3_double()
    refl["imageset_id"] = flex.int(len(refl))
    refl["xyzobs.mm.value"] = flex.vec3_double(len(refl))

    d = {}
    d["crystal_symmetry"] = cs
    d["reflections"] = refl
    return d


@pytest.mark.parametrize(
    "histogram_binning,nearest_neighbor_percentile", [("linear", None), ("log", 0.99)]
)
def test_max_cell(setup, histogram_binning, nearest_neighbor_percentile):
    reflections = setup["reflections"]
    crystal_symmetry = setup["crystal_symmetry"]

    max_cell_multiplier = 1.3
    max_cell = find_max_cell(
        reflections,
        max_cell_multiplier=max_cell_multiplier,
        histogram_binning=histogram_binning,
        nearest_neighbor_percentile=nearest_neighbor_percentile,
    )

    known_max_cell = max(
        crystal_symmetry.primitive_setting().unit_cell().parameters()[:3]
    )
    assert max_cell.max_cell > known_max_cell


def test_max_cell_low_res_with_high_res_noise(setup):
    reflections = setup["reflections"]
    crystal_symmetry = setup["crystal_symmetry"]

    rlp = reflections["rlp"]
    # select only low resolution reflections
    reflections = reflections.select(1 / rlp.norms() > 4)

    n = int(0.1 * reflections.size())
    rlp_noise = flex.vec3_double(*(flex.random_double(n) for i in range(3)))
    reflections["rlp"].extend(rlp_noise)
    reflections["imageset_id"].extend(flex.int(rlp_noise.size()))
    reflections["xyzobs.mm.value"].extend(flex.vec3_double(rlp_noise.size()))

    max_cell_multiplier = 1.3
    max_cell = find_max_cell(reflections, max_cell_multiplier=max_cell_multiplier)

    known_max_cell = max(
        crystal_symmetry.primitive_setting().unit_cell().parameters()[:3]
    )
    assert max_cell.max_cell > known_max_cell



"""
This program runs indexing on the spotfinding results from a
still sequence i.e. SSX data. This wraps a call to the regular
indexing code, and so all parameters from dials.index can be set.

If a unit cell is given, indexing of each image will be attempted with the
fft1d algorithm, followed by the real_space_grid_search algorithm if indexing
with the fft1d algorithm was unsuccessful. If no unit cell is given, only fft1d
indexing can be attempted.

Indexing statistics are reported in a table and a unit cell clustering analysis
is performed, which can be useful for assessing the crystal symmetry if the unit
cell is unknown. An extensive html report is generated showing the indexing
and clustering statistics. The indexed data are saved into a single reflection
file and a single experiment list file, with a joint detector and beam model.

Further program documentation can be found at dials.github.io/ssx_processing_guide.html

Usage:
    dials.ssx_index imported.expt strong.refl
    dials.ssx_index imported.expt strong.refl unit_cell=x space_group=y
"""

from __future__ import annotations

import json
import logging
import sys
import time
from functools import reduce

from cctbx import crystal
from libtbx import Auto, phil
from libtbx.introspection import number_of_processors

from dials.algorithms.indexing.ssx.analysis import (
    generate_html_report,
    generate_plots,
    make_summary_table,
    report_on_crystal_clusters,
)
from dials.algorithms.indexing.ssx.processing import index
from dials.util import log, show_mail_handle_errors
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

try:
    from typing import List
except ImportError:
    pass

logger = logging.getLogger("dials")

program_defaults_phil_str = """
indexing {
  method = fft1d
  stills {
    indexer = stills
  }
}
output.log = dials.ssx_index.log
refinement {
  parameterisation {
    auto_reduction {
      min_nref_per_parameter = 1
      action = fix
    }
    beam.fix = all
    detector.fix = all
    scan_varying = False
  }
  reflections {
    weighting_strategy.override = stills
    outlier.algorithm = null
  }
}
"""

phil_scope = phil.parse(
    """
method = *fft1d *real_space_grid_search
    .type = choice(multi=True)
nproc = Auto
    .type = int
    .expert_level = 1
    .help = "Set the number of processors to use in indexing"
min_spots = 10
    .type = int
    .expert_level = 2
    .help = "Images with fewer than this number of strong spots will not be indexed"
output.html = dials.ssx_index.html
    .type = str
output.json = None
    .type = str
output.nuggets = None
    .type = path
    .help = "Specify a directory to which a per-image summary json will be saved"
            "during processing, as each image is indexed, to enable live monitoring."
include scope dials.command_line.index.phil_scope
""",
    process_includes=True,
).fetch(phil.parse(program_defaults_phil_str))

phil_scope.adopt_scope(
    phil.parse(
        """
    individual_log_verbosity = 1
    .type =int
"""
    )
)


@show_mail_handle_errors()
def run(args: List[str] = None, phil: phil.scope = phil_scope) -> None:
    """
    Run dials.ssx_index as from the command line.

    This program takes an imported experiment list and a reflection table
    of strong spots and performs parallelised indexing for synchrotron
    serial crystallography experiments. This is done by calling the regular
    dials indexing code and capturing output to provide a html report, and
    outputs a multi-image indexed.expt and indexed.refl file containing the
    indexed data.
    """

    parser = ArgumentParser(
        usage="dials.ssx_index imported.expt strong.refl [options]",
        read_experiments=True,
        read_reflections=True,
        phil=phil_scope,
        check_format=False,
        epilog=__doc__,
    )
    params, options = parser.parse_args(args=args, show_diff_phil=False)

    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        sys.exit()

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    log.config(verbosity=options.verbose, logfile=params.output.log)
    params.individual_log_verbosity = options.verbose
    logger.info(dials_version())

    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    if params.nproc is Auto:
        params.nproc = number_of_processors(return_value_if_unknown=1)

    if params.nproc > 1:
        params.indexing.nproc = params.nproc

    logger.info(f"Using {params.indexing.nproc} processes for indexing")

    st = time.time()
    indexed_experiments, indexed_reflections, summary_data = index(
        experiments, reflections[0], params
    )

    summary_table = make_summary_table(summary_data)
    logger.info("\nSummary of images sucessfully indexed\n" + summary_table)

    n_images = reduce(
        lambda a, v: a + (v[0]["n_indexed"] > 0), summary_data.values(), 0
    )
    logger.info(f"{indexed_reflections.size()} spots indexed on {n_images} images\n")

    crystal_symmetries = [
        crystal.symmetry(
            unit_cell=expt.crystal.get_unit_cell(),
            space_group=expt.crystal.get_space_group(),
        )
        for expt in indexed_experiments
    ]
    if crystal_symmetries:
        cluster_plots, _ = report_on_crystal_clusters(
            crystal_symmetries,
            make_plots=(params.output.html or params.output.json),
        )

    logger.info(f"Saving indexed experiments to {params.output.experiments}")
    indexed_experiments.as_file(params.output.experiments)
    logger.info(f"Saving indexed reflections to {params.output.reflections}")
    indexed_reflections.as_file(params.output.reflections)

    if (params.output.html or params.output.json) and indexed_experiments:
        summary_plots = generate_plots(summary_data)
        if cluster_plots:
            summary_plots.update(cluster_plots)
        if params.output.html:
            generate_html_report(summary_plots, params.output.html)
        if params.output.json:
            with open(params.output.json, "w") as outfile:
                json.dump(summary_plots, outfile)

    logger.info(f"Total time: {time.time() - st:.2f}s")
    logger.info(
        "Further program documentation can be found at dials.github.io/ssx_processing_guide.html"
    )


if __name__ == "__main__":
    run()



"""Algorithms for determination of Laue group symmetry."""

from __future__ import annotations

import json
import logging
import math

import scipy.stats

import libtbx
from cctbx import crystal, sgtbx
from scitbx.array_family import flex
from scitbx.math import five_number_summary

import dials.util
from dials.algorithms.symmetry import symmetry_base

logger = logging.getLogger(__name__)


class LaueGroupAnalysis(symmetry_base):
    """Determination of Laue group symmetry using algorithms similar to POINTLESS.

    See also:
      `Evans, P. (2006). Acta Cryst. D62, 72-82
      <https://doi.org/10.1107/S0907444905036693>`_ and
      `Evans, P. R. (2011). Acta Cryst. D67, 282-292
      <https://doi.org/10.1107/S090744491003982X>`_.
    """

    def __init__(
        self,
        intensities,
        normalisation="ml_aniso",
        lattice_symmetry_max_delta=2.0,
        d_min=libtbx.Auto,
        min_i_mean_over_sigma_mean=4,
        min_cc_half=0.6,
        relative_length_tolerance=None,
        absolute_angle_tolerance=None,
        best_monoclinic_beta=True,
    ):
        """Initialise a LaueGroupAnalysis object.

        Args:
          intensities (cctbx.miller.array): The intensities on which to perform
            symmetry analysis.
          normalisation (str): The normalisation method to use. Possible choices are
            'kernel', 'quasi', 'ml_iso' and 'ml_aniso'. Set to None to switch off
            normalisation altogether.
          lattice_symmetry_max_delta (float): The maximum value of delta for
            determining the lattice symmetry using the algorithm of Le Page (1982).
          d_min (float): Optional resolution cutoff to be applied to the input
            intensities. If set to :data:`libtbx.Auto` then d_min will be
            automatically determined according to the parameters
            ``min_i_mean_over_sigma_mean`` and ``min_cc_half``.
          min_i_mean_over_sigma_mean (float): minimum value of :math:`|I|/|sigma(i)|` for
            automatic determination of resolution cutoff.
          min_cc_half (float): minimum value of CC1/2 for automatic determination of
            resolution cutoff.
          relative_length_tolerance (float): Relative length tolerance in checking
            consistency of input unit cells against the median unit cell.
          absolute_angle_tolerance (float): Absolute angle tolerance in checking
            consistency of input unit cells against the median unit cell.
          best_monoclinic_beta (bool): If True, then for monoclinic centered cells, I2
            will be preferred over C2 if it gives a less oblique cell (i.e. smaller
            beta angle).
        """
        super().__init__(
            intensities,
            normalisation=normalisation,
            lattice_symmetry_max_delta=lattice_symmetry_max_delta,
            d_min=d_min,
            min_i_mean_over_sigma_mean=min_i_mean_over_sigma_mean,
            min_cc_half=min_cc_half,
            relative_length_tolerance=relative_length_tolerance,
            absolute_angle_tolerance=absolute_angle_tolerance,
            best_monoclinic_beta=best_monoclinic_beta,
        )

        self._estimate_cc_sig_fac()
        self._estimate_cc_true()
        self._score_symmetry_elements()
        self._score_laue_groups()

    def _estimate_cc_sig_fac(self):
        """Estimation of sigma(CC) as a function of sample size.

        Estimate the error in the correlation coefficient, sigma(CC) by using
        pairs of reflections at similar resolutions that are not related by
        potential symmetry. Using pairs of unrelated reflections at similar
        resolutions, calculate sigma(CC) == rms(CC) for groups of size N = 3..200.
        The constant CCsigFac is obtained from a linear fit of
        sigma(CC) to 1/N^(1/2), i.e.:
            sigma(CC) = CCsigFac/N^(1/2)
        """

        max_bins = 500
        reflections_per_bin = max(
            200, int(math.ceil(self.intensities.size() / max_bins))
        )
        binner = self.intensities.setup_binner_counting_sorted(
            reflections_per_bin=reflections_per_bin
        )

        a = flex.double()
        b = flex.double()
        ma_tmp = self.intensities.customized_copy(
            crystal_symmetry=crystal.symmetry(
                space_group=self.lattice_group,
                unit_cell=self.intensities.unit_cell(),
                assert_is_compatible_unit_cell=False,
            )
        ).map_to_asu()
        for i in range(binner.n_bins_all()):
            count = binner.counts()[i]
            if count == 0:
                continue
            bin_isel = binner.array_indices(i)
            p = flex.random_permutation(count)
            p = p[: 2 * (count // 2)]  # ensure even count
            ma_a = ma_tmp.select(bin_isel.select(p[: count // 2]))
            ma_b = ma_tmp.select(bin_isel.select(p[count // 2 :]))
            # only choose pairs of reflections that don't have the same indices
            # in the asu of the lattice group
            sel = ma_a.indices() != ma_b.indices()
            a.extend(ma_a.data().select(sel))
            b.extend(ma_b.data().select(sel))

        perm = flex.random_selection(a.size(), min(20000, a.size()))
        a = a.select(perm)
        b = b.select(perm)

        self.corr_unrelated = CorrelationCoefficientAccumulator(a, b)

        n_pairs = a.size()
        min_num_groups = 10  # minimum number of groups
        max_n_group = int(min(n_pairs / min_num_groups, 200))  # maximum number in group
        min_n_group = int(min(5, max_n_group))  # minimum number in group

        if (max_n_group - min_n_group) < 4:
            self.cc_sig_fac = 0
            return

        mean_ccs = flex.double()
        rms_ccs = flex.double()
        ns = flex.double()
        for n in range(min_n_group, max_n_group + 1):
            ns.append(n)
            ccs = flex.double()
            for i in range(200):
                isel = flex.random_selection(a.size(), n)
                corr = CorrelationCoefficientAccumulator(a.select(isel), b.select(isel))
                ccs.append(corr.coefficient())

            mean_ccs.append(flex.mean(ccs))
            rms_ccs.append(flex.mean(flex.pow2(ccs)) ** 0.5)

        x = 1 / flex.pow(ns, 0.5)
        y = rms_ccs
        fit = flex.linear_regression(x, y)

        if fit.is_well_defined():
            self.cc_sig_fac = fit.slope()
        else:
            self.cc_sig_fac = 0

    def _estimate_cc_true(self):

        # A1.2. Estimation of E(CC; S).

        # (i)

        var_intensities = flex.mean_and_variance(
            self.intensities.data()
        ).unweighted_sample_variance()
        var_sigmas = flex.mean_and_variance(flex.pow2(self.intensities.sigmas())).mean()
        self.E_cc_true = var_intensities / (var_intensities + var_sigmas)

        # (ii)

        reindexed_intensities = self.intensities.change_basis(
            sgtbx.change_of_basis_op("-x,-y,-z")
        ).map_to_asu()
        x, y = self.intensities.common_sets(
            reindexed_intensities, assert_is_similar_symmetry=False
        )
        self.cc_identity = CorrelationCoefficientAccumulator(x.data(), y.data())

        min_sd = 0.05
        min_sample = 10
        sigma_1 = max(min_sd, self.cc_sig_fac / 200**0.5)
        w1 = 0
        w2 = 0
        if sigma_1 > 0.0001:
            w1 = 1 / sigma_1**2
        if self.cc_identity.n() > min_sample:
            sigma_2 = max(min_sd, self.cc_sig_fac / self.cc_identity.n() ** 0.5)
            w2 = 1 / sigma_2**2

        assert (w1 + w2) > 0
        self.cc_true = (w1 * self.E_cc_true + w2 * self.cc_identity.coefficient()) / (
            w1 + w2
        )

        logger.debug("cc_true = w1 * E_cc_true + w2 * cc_identity)/(w1 + w2)")
        logger.debug("w1: %g", w1)
        logger.debug("w2: %g", w2)
        logger.debug("E_cc_true: %g", self.E_cc_true)
        logger.debug("cc_identity: %g", self.cc_identity.coefficient())
        logger.debug("cc_true: %g", self.cc_true)

    def _score_symmetry_elements(self):
        self.sym_op_scores = []
        for smx in self.lattice_group.smx():
            if smx.r().info().sense() < 0:
                continue
            self.sym_op_scores.append(
                ScoreSymmetryElement(
                    self.intensities, smx, self.cc_true, self.cc_sig_fac
                )
            )

    def _score_laue_groups(self):
        subgroup_scores = [
            ScoreSubGroup(subgrp, self.sym_op_scores)
            for subgrp in self.subgroups.result_groups
        ]
        total_likelihood = sum(score.likelihood for score in subgroup_scores)
        for score in subgroup_scores:
            score.likelihood /= total_likelihood
        self.subgroup_scores = sorted(
            subgroup_scores, key=lambda score: score.likelihood, reverse=True
        )

        # The 'confidence' scores are derived from the total probability of the best
        # solution p_best and that for the next best solution p_next:
        #   confidence = [p_best * (p_best - p_next)]^1/2.

        for i, score in enumerate(self.subgroup_scores[:-1]):
            next_score = self.subgroup_scores[i + 1]
            if score.likelihood > 0 and next_score.likelihood > 0:
                lgc = score.likelihood * (score.likelihood - next_score.likelihood)
                confidence = abs(lgc) ** 0.5
                if lgc < 0:
                    confidence = -confidence
                score.confidence = confidence

        self.best_solution = self.subgroup_scores[0]

    def __str__(self):
        """Return a string representation of the results.

        Returns:
          str:
        """
        output = []
        output.append("Input crystal symmetry:")
        output.append(
            str(
                self.input_intensities[0]
                .crystal_symmetry()
                .customized_copy(unit_cell=self.median_unit_cell)
            )
        )
        output.append(f"Change of basis op to minimum cell: {self.cb_op_inp_min}")
        output.append("Crystal symmetry in minimum cell:")
        output.append(str(self.intensities.crystal_symmetry()))
        output.append(f"Lattice point group: {self.lattice_group.info()}")
        output.append(
            "\nOverall CC for %i unrelated pairs: %.3f"
            % (self.corr_unrelated.n(), self.corr_unrelated.coefficient())
        )
        output.append(
            "Estimated expectation value of true correlation coefficient E(CC) = %.3f"
            % self.E_cc_true
        )
        if self.cc_sig_fac:
            output.append(f"Estimated sd(CC) = {self.cc_sig_fac:.3f} / sqrt(N)")
        else:
            output.append("Too few reflections to estimate sd(CC).")
        output.append(
            "Estimated E(CC) of true correlation coefficient from identity = %.3f"
            % self.cc_true
        )

        header = ("likelihood", "Z-CC", "CC", "N", "", "Operator")
        rows = [header]
        for score in self.sym_op_scores:
            if score.likelihood > 0.9:
                stars = "***"
            elif score.likelihood > 0.7:
                stars = "**"
            elif score.likelihood > 0.5:
                stars = "*"
            else:
                stars = ""
            rows.append(
                (
                    f"{score.likelihood:.3f}",
                    f"{score.z_cc:.2f}",
                    f"{score.cc.coefficient():.2f}",
                    "%i" % score.n_refs,
                    stars,
                    f"{score.sym_op.r().info()}",
                )
            )
        output.append("\n" + "-" * 80 + "\n")
        output.append("Scoring individual symmetry elements\n")
        output.append(dials.util.tabulate(rows, headers="firstrow"))

        header = (
            "Patterson group",
            "",
            "Likelihood",
            "NetZcc",
            "Zcc+",
            "Zcc-",
            "CC",
            "CC-",
            "delta",
            "Reindex operator",
        )
        rows = [header]
        for score in self.subgroup_scores:
            if score.likelihood > 0.8:
                stars = "***"
            elif score.likelihood > 0.6:
                stars = "**"
            elif score.likelihood > 0.4:
                stars = "*"
            else:
                stars = ""
            rows.append(
                (
                    f"{score.subgroup['best_subsym'].space_group_info()}",
                    stars,
                    f"{score.likelihood:.3f}",
                    f"{score.z_cc_net: .2f}",
                    f"{score.z_cc_for: .2f}",
                    f"{score.z_cc_against: .2f}",
                    f"{score.cc_for.coefficient(): .2f}",
                    f"{score.cc_against.coefficient(): .2f}",
                    f"{score.subgroup['max_angular_difference']:.1f}",
                    f"{score.subgroup['cb_op_inp_best']}",
                )
            )
        output.append("\n" + "-" * 80 + "\n")
        output.append("Scoring all possible sub-groups\n")
        output.append(dials.util.tabulate(rows, headers="firstrow"))

        output.append(
            "\nBest solution: %s"
            % self.best_solution.subgroup["best_subsym"].space_group_info()
        )
        output.append(
            f"Unit cell: {self.best_solution.subgroup['best_subsym'].unit_cell()}"
        )
        output.append(
            f"Reindex operator: {self.best_solution.subgroup['cb_op_inp_best']}"
        )
        output.append(f"Laue group probability: {self.best_solution.likelihood:.3f}")
        output.append(f"Laue group confidence: {self.best_solution.confidence:.3f}\n")
        return "\n".join(output)

    def as_dict(self):
        """Return a dictionary representation of the results.

        Returns:
          dict
        """
        d = {
            "input_symmetry": {
                "hall_symbol": self.input_intensities[0]
                .space_group()
                .type()
                .hall_symbol(),
                "unit_cell": self.median_unit_cell.parameters(),
            },
            "cb_op_inp_min": self.cb_op_inp_min.as_xyz(),
            "lattice_point_group": self.lattice_group.type().hall_symbol(),
            "cc_unrelated_pairs": self.corr_unrelated.coefficient(),
            "n_unrelated_pairs": self.corr_unrelated.n(),
            "E_cc_true": self.E_cc_true,
            "cc_sig_fac": self.cc_sig_fac,
            "cc_true": self.cc_true,
        }

        d["sym_op_scores"] = [score.as_dict() for score in self.sym_op_scores]
        d["subgroup_scores"] = [score.as_dict() for score in self.subgroup_scores]
        return d

    def as_json(self, filename=None, indent=2):
        """Return a json representation of the results.

        Args:
          filename (str): Optional filename to export the json representation of
            the results.
          indent (int): The indent level for pretty-printing of the json. If ``None``
            is the most compact representation.

        Returns:
          str:
        """
        json_str = json.dumps(self.as_dict(), indent=indent)
        if filename:
            with open(filename, "w") as f:
                f.write(json_str)
        return json_str


class ScoreCorrelationCoefficient:
    def __init__(self, cc, sigma_cc, expected_cc, lower_bound=-1, upper_bound=1, k=2):
        self.cc = cc
        self.sigma_cc = sigma_cc
        self.expected_cc = expected_cc
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._k = k
        self._compute_p_cc_given_s()
        self._compute_p_cc_given_not_s()

    def _compute_p_cc_given_s(self):
        self._p_cc_given_s = trunccauchy_pdf(
            self.cc,
            self._lower_bound,
            self._upper_bound,
            self.expected_cc,
            self.sigma_cc,
        )

    def _compute_p_cc_given_not_s(self):
        sump = scipy.integrate.quad(self._numerator, 0, 1)[0]
        sumw = scipy.integrate.quad(self._denominator, 0, 1)[0]
        self._p_cc_given_not_s = sump / sumw

    @property
    def p_cc_given_s(self):
        """Probability of observing this CC if the sym op is present, p(CC; S).

        Modelled by a Cauchy distribution centred on cc_true and width gamma = sigma_cc
        """
        return self._p_cc_given_s

    @property
    def p_cc_given_not_s(self):
        """Probability of observing this CC if the sym op is NOT present, p(CC; !S)."""
        return self._p_cc_given_not_s

    @property
    def p_s_given_cc(self):
        """The likelihood of this symmetry element being present.

        p(S; CC) = p(CC; S) / (p(CC; S) + p(CC; !S))
        """
        return self._p_cc_given_s / (self._p_cc_given_s + self._p_cc_given_not_s)

    def _p_mu_power_pdf(self, m):
        return (1.0 - pow(m, self._k)) ** (1.0 / self._k)

    def _numerator(self, x):
        return trunccauchy_pdf(
            self.cc, self._lower_bound, self._upper_bound, loc=x, scale=self.sigma_cc
        ) * self._p_mu_power_pdf(x)

    def _denominator(self, x):
        return self._p_mu_power_pdf(x)


class ScoreSymmetryElement:
    """Analyse intensities for presence of a given symmetry operation.

    1) Calculate the correlation coefficient, CC, for the given sym op.

    2) Calculate the probability of observing this CC if the sym op is present,
       p(CC; S), modelled by a Cauchy distribution centred on cc_true and width
       gamma = sigma_cc.

    3) Calculate the probability of observing this CC if the sym op is
       NOT present, p(CC; !S).

    4) Calculate the likelihood of symmetry element being present,
       p(S; CC) = p(CC; S) / (p(CC; S) + p(CC; !S))

    See appendix A1 of `Evans, P. R. (2011). Acta Cryst. D67, 282-292.
    <https://doi.org/10.1107/S090744491003982X>`_
    """

    def __init__(self, intensities, sym_op, cc_true, cc_sig_fac):
        """Initialise a ScoreSymmetryElement object.

        Args:
          intensities (cctbx.miller.array): The intensities on which to perform
            symmetry analysis.
          sym_op (cctbx.sgtbx.rt_mx): The symmetry operation for analysis.
          cc_true (float): the expected value of CC if the symmetry element is present,
            E(CC; S)
          cc_sig_fac (float): Estimation of sigma(CC) as a function of sample size.
        """
        self.sym_op = sym_op
        assert self.sym_op.r().info().sense() >= 0
        self.cc = CorrelationCoefficientAccumulator()
        cb_op = sgtbx.change_of_basis_op(self.sym_op)
        cb_ops = [cb_op]
        if self.sym_op.r().order() > 2:
            # include inverse symmetry operation
            cb_ops.append(cb_op.inverse())
        for cb_op in cb_ops:
            if cb_op.is_identity_op():
                cb_op = sgtbx.change_of_basis_op("-x,-y,-z")
            reindexed_intensities = intensities.change_basis(cb_op).map_to_asu()
            x, y = intensities.common_sets(
                reindexed_intensities, assert_is_similar_symmetry=False
            )
            sel = sgtbx.space_group().expand_smx(self.sym_op).epsilon(x.indices()) == 1
            x = x.select(sel)
            y = y.select(sel)

            outliers = flex.bool(len(x.data()), False)
            iqr_multiplier = 20  # very generous tolerance
            for col in (x.data(), y.data()):
                if col.size():
                    min_x, q1_x, med_x, q3_x, max_x = five_number_summary(col)
                    iqr_x = q3_x - q1_x
                    cut_x = iqr_multiplier * iqr_x
                    outliers.set_selected(col > q3_x + cut_x, True)
                    outliers.set_selected(col < q1_x - cut_x, True)
            if outliers.count(True):
                logger.debug(
                    "Rejecting %s outlier value%s",
                    libtbx.utils.plural_s(outliers.count(True)),
                )
                x = x.select(~outliers)
                y = y.select(~outliers)

            self.cc += CorrelationCoefficientAccumulator(x.data(), y.data())

        self.n_refs = self.cc.n()
        if self.n_refs <= 0:
            self.likelihood = 0
            self.z_cc = 0
            return

        self.sigma_cc = max(0.1, cc_sig_fac / self.n_refs**0.5)
        self.z_cc = self.cc.coefficient() / self.sigma_cc
        score_cc = ScoreCorrelationCoefficient(
            self.cc.coefficient(), self.sigma_cc, cc_true
        )
        self.p_cc_given_s = score_cc.p_cc_given_s
        self.p_cc_given_not_s = score_cc.p_cc_given_not_s
        self.likelihood = score_cc.p_s_given_cc

    def __str__(self):
        """Return a string representation of the symmetry element scoring.

        Returns:
          str:
        """
        return "%.3f %.2f %.2f %i %s" % (
            self.likelihood,
            self.z_cc,
            self.cc.coefficient(),
            self.n_refs,
            self.sym_op.r().info(),
        )

    def as_dict(self):
        """Return a dictionary representation of the symmetry element scoring.

        The dictionary will contain the following keys:
          - likelihood: The likelihood of the symmetry element being present
          - z_cc: The Z-score for the correlation coefficient
          - cc: The correlation coefficient for the symmetry element
          - n_ref: The number of reflections contributing to the correlation
            coefficient
          - operator: The xyz representation of the symmetry element

        Returns:
          dict:
        """
        return {
            "likelihood": self.likelihood,
            "z_cc": self.z_cc,
            "cc": self.cc.coefficient(),
            "n_ref": self.n_refs,
            "operator": self.sym_op.as_xyz(),
        }


class ScoreSubGroup:
    """Score the probability of a given subgroup being the true subgroup.

    1) Calculates the combined correlation coefficients for symmetry operations
       present/absent from the subgroup.

    2) Calculates overall Zcc scores for symmetry elements present/absent from
       the subgroup.

    3) Calculates the overall likelihood for this subgroup.

    See appendix A2 of `Evans, P. R. (2011). Acta Cryst. D67, 282-292.
    <https://doi.org/10.1107/S090744491003982X>`_
    """

    def __init__(self, subgroup, sym_op_scores):
        """Initialise a ScoreSubGroup object.

        Args:
          subgroup (dict): A dictionary describing the subgroup as generated by
            :class:`cctbx.sgtbx.lattice_symmetry.metric_subgroups`.
          sym_op_scores (list): A list of :class:`ScoreSymmetryElement` objects for each
            symmetry element possibly in the lattice symmetry.
        """
        # Combined correlation coefficients for symmetry operations
        # present/absent from subgroup
        self.subgroup = subgroup
        patterson_group = subgroup["subsym"].space_group()
        self.cc_for = CorrelationCoefficientAccumulator()
        self.cc_against = CorrelationCoefficientAccumulator()
        for score in sym_op_scores:
            if score.sym_op in patterson_group:
                self.cc_for += score.cc
            else:
                self.cc_against += score.cc

        # Overall Zcc scores for symmetry elements present/absent from subgroup
        self.z_cc_for = 0
        self.z_cc_against = 0
        n_for = 0
        n_against = 0
        PL_for = 0
        PL_against = 0
        power = 2
        for score in sym_op_scores:
            if score.n_refs <= 2:
                continue
            if score.sym_op in patterson_group:
                self.z_cc_for += score.z_cc**power
                n_for += 1
                PL_for += math.log(score.p_cc_given_s)
            else:
                self.z_cc_against += score.z_cc**power
                n_against += 1
                PL_against += math.log(score.p_cc_given_not_s)

        # Overall likelihood for this subgroup
        self.likelihood = math.exp(PL_for + PL_against)

        if n_against > 0:
            self.z_cc_against = (self.z_cc_against / n_against) ** (1 / power)
        if n_for > 0:
            self.z_cc_for = (self.z_cc_for / n_for) ** (1 / power)
        self.z_cc_net = self.z_cc_for - self.z_cc_against
        self.confidence = 0

    def __str__(self):
        """Return a string representation of the subgroup scores.

        Returns:
          str:
        """
        return "{} {:.3f} {:.2f} {:.2f} {:.2f} {:.2f} {:.2f}".format(
            self.subgroup["best_subsym"].space_group_info(),
            self.likelihood,
            self.z_cc_net,
            self.z_cc_for,
            self.z_cc_against,
            self.cc_for.coefficient(),
            self.cc_against.coefficient(),
        )

    def as_dict(self):
        """Return a dictionary representation of the subgroup scoring.

        The dictionary will contain the following keys:
          - patterson_group: The current subgroup
          - likelihood: The likelihood of the subgroup being correct
          - confidence: The confidence of the subgroup being correct
          - z_cc_for: The combined Z-scores for all symmetry elements present in the
            subgroup
          - z_cc_against: The combined Z-scores for all symmetry elements present in
            the lattice group but not in the subgroup
          - z_cc_net: The net Z-score, i.e. z_cc_for - z_cc_against
          - cc_for: The overall correlation coefficient for all symmetry elements
            present in the subgroup
          - cc_against: The overall correlation coefficient for all symmetry
            elements present in the lattice group but not in the subgroup
          - max_angular_difference: The maximum angular difference between the
            symmetrised unit cell and the P1 unit cell.
          - cb_op: The change of basis operation from the input unit cell to the
            'best' unit cell.

        Returns:
          dict:
        """
        return {
            "patterson_group": self.subgroup["best_subsym"]
            .space_group()
            .type()
            .hall_symbol(),
            "likelihood": self.likelihood,
            "confidence": self.confidence,
            "z_cc_net": self.z_cc_net,
            "z_cc_for": self.z_cc_for,
            "z_cc_against": self.z_cc_against,
            "cc_for": self.cc_for.coefficient(),
            "cc_against": self.cc_against.coefficient(),
            "max_angular_difference": self.subgroup["max_angular_difference"],
            "cb_op": f"{self.subgroup['cb_op_inp_best']}",
        }


class CorrelationCoefficientAccumulator:
    """Class for incremental computation of correlation coefficients.

    Uses the single-pass formula for Pearson correlation coefficient:
      https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#For_a_sample
    """

    def __init__(self, x=None, y=None):
        """Initialise a CorrelationCoefficientAccumulator object.

        Args:
          x (list): Optional list of `x` values to initialise the accumulator.
          y (list): Optional list of `y` values to initialise the accumulator.
        """
        self._n = 0
        self._sum_x = 0
        self._sum_y = 0
        self._sum_xy = 0
        self._sum_x_sq = 0
        self._sum_y_sq = 0
        if x is not None and y is not None:
            self.accumulate(x, y)

    def accumulate(self, x, y):
        """Accumulate the `x` and `y` values provided.

        Args:
          x (list): The list of `x` values to accumulate.
          y (list): The list of `y` values to accumulate.
        """
        assert x.size() == y.size()
        self._n += x.size()
        self._sum_x += flex.sum(x)
        self._sum_y += flex.sum(y)
        self._sum_xy += flex.sum(x * y)
        self._sum_x_sq += flex.sum(flex.pow2(x))
        self._sum_y_sq += flex.sum(flex.pow2(y))

    def coefficient(self):
        """Calculate the correlation coefficient.

        Returns:
          float: The correlation coefficient.
        """
        if self._n == 0:
            return 0
        return self.numerator() / self.denominator()

    def n(self):
        """Return the number of values contributing to the correlation coefficient.

        Returns:
          n (int)
        """
        return self._n

    def numerator(self):
        r"""Calculate the numerator of the correlation coefficient formula.

        .. math:: n \sum{x y} - \sum{x} \sum{y}

        Returns:
          float: The value of the numerator.
        """
        return self._n * self._sum_xy - self._sum_x * self._sum_y

    def denominator(self):
        r"""Calculate the denominator of the correlation coefficient formula.

        .. math:: \sqrt{n \sum{x^2} - \sum{x}^2} \sqrt{n \sum{y^2} - \sum{y}^2}

        Returns:
          float: The value of the denominator.
        """
        return math.sqrt(self._n * self._sum_x_sq - self._sum_x**2) * math.sqrt(
            self._n * self._sum_y_sq - self._sum_y**2
        )

    def __iadd__(self, other):
        """Add together two instances of :class:`CorrelationCoefficientAccumulator`.

        Args:
          other (CorrelationCoefficientAccumulator):
            The :class:`CorrelationCoefficientAccumualator` to add to the current object.

        Returns:
          self (CorrelationCoefficientAccumulator): The current object.
        """
        self._n += other._n
        self._sum_x += other._sum_x
        self._sum_y += other._sum_y
        self._sum_xy += other._sum_xy
        self._sum_x_sq += other._sum_x_sq
        self._sum_y_sq += other._sum_y_sq
        return self


def trunccauchy_pdf(x, a, b, loc=0, scale=1):
    """Calculate a truncated Cauchy probability density function.

    Args:
      x (float): The point at which to calculate the PDF.
      a (float): The lower bound of the truncated distribution.
      b (float): The upper bound of the truncated distribution.
      loc (float): The location parameter for the Cauchy distribution.
      scale (float): The scale parameter for the Cauchy distribution.

    Returns:
      float: The value of the probability density function.
    """
    assert b > a
    rv = scipy.stats.cauchy(loc=loc, scale=scale)
    return rv.pdf(x) / (rv.cdf(b) - rv.cdf(a))
